# -*- coding: utf-8 -*-

'''Text block objects based on PDF raw dict extracted with ``PyMuPDF``.

Data structure based on this `link <https://pymupdf.readthedocs.io/en/latest/textpage.html>`_::

    {
        # raw dict
        # --------------------------------
        'type': 0,
        'bbox': (x0,y0,x1,y1),
        'lines': [ lines ]

        # introduced dict
        # --------------------------------
        'before_space': bs,
        'after_space': as,
        'line_space': ls,

        'alignment': 0,
        'left_space': 10.0,
        'right_space': 0.0,

        'tab_stops': [15.4, 35.0]
    }
'''

from docx.shared import (Pt,Inches)
from docx.enum.text import WD_ALIGN_PARAGRAPH
from .Lines import Lines
from ..image.ImageSpan import ImageSpan
from ..common.share import (RectType, TextAlignment, lower_round)
from ..common.Block import Block
from ..common.share import (rgb_component_from_name, lower_round)
from ..common import constants
from ..common import docx


class TextBlock(Block):
    '''Text block.'''
    def __init__(self, raw:dict=None):
        raw = raw or {}
        
        # remove key 'bbox' since it is calculated from contained lines
        if 'bbox' in raw: raw.pop('bbox') 
        super().__init__(raw)

        # collect lines
        self.lines = Lines(parent=self).restore(raw.get('lines', []))

        # set type
        self.set_text_block()


    @property
    def text(self):
        '''Text content in block. Note image is counted as a placeholder ``<image>``.'''
        lines_text = [line.text for line in self.lines]
        return ''.join(lines_text)

    @property
    def raw_text(self):
        '''Raw text content in block without considering images.'''
        lines_text = [line.raw_text for line in self.lines]
        return ''.join(lines_text)

    @property
    def white_space_only(self):
        '''If this block contains only white space or not. If True, this block is safe to be removed.'''
        return all(line.white_space_only for line in self.lines)

    @property
    def text_direction(self):
        '''All lines contained in text block must have same text direction. 
        Otherwise, set normal direction.
        '''            
        return self.lines.text_direction

    @property
    def average_row_gap(self):
        '''Average distance between adjacent two physical rows.'''
        idx = 1 if self.is_horizontal_text else 0
        rows = self.lines.group_by_physical_rows()
        num = len(rows)

        # no gap if single row
        if num==1: return None
        
        # multi-lines block
        block_height = self.bbox[idx+2]-self.bbox[idx]
        f_max_row_height = lambda row: max(abs(line.bbox[idx+2]-line.bbox[idx]) for line in row)
        sum_row_height = sum(map(f_max_row_height, rows))
        return (block_height-sum_row_height) / (num-1)
    

    @property
    def row_count(self):
        '''Count of physical rows.'''
        return len(self.lines.group_by_physical_rows())


    def store(self):
        res = super().store()
        res.update({
            'lines': self.lines.store()
        })
        return res


    def add(self, line_or_lines):
        '''Add line or lines to TextBlock.'''        
        if isinstance(line_or_lines, (Lines, list, tuple)):
            for line in line_or_lines:
                self.lines.append(line)
        else:
            self.lines.append(line_or_lines)


    def plot(self, page):
        '''Plot block/line/span area for debug purpose.
        
        Args:
            page (fitz.Page): pdf page.
        '''
        # block border in blue
        blue = rgb_component_from_name('blue')   
        super().plot(page, stroke=blue, dashes='[3.0 3.0] 0')

        # lines and spans
        for line in self.lines:
            # line border in red
            red = rgb_component_from_name('red')
            line.plot(page, stroke=red)

            # span regions in random color
            for span in line.spans:
                c = rgb_component_from_name('')                
                span.plot(page, color=c)


    def parse_text_format(self, shapes):
        '''Parse text format with style represented by rectangles.
        
        Args:
            shapes (Shapes): Shapes representing potential styles applied on blocks.
        '''
        flag = False

        # use each rectangle (a specific text format) to split line spans
        for shape in shapes:

            # a same style shape applies on only one block
            # EXCEPTION: hyperlink shape is determined in advance
            if not shape.equal_to_type(RectType.HYPERLINK) and shape.is_determined: continue

            # any intersection with current block?
            if not self.bbox.intersects(shape.bbox): continue

            # yes, then go further to lines in block
            if self.lines.parse_text_format(shape):
                flag = True

        return flag


    def parse_horizontal_spacing(self, bbox,
                    line_separate_threshold:float,
                    line_break_width_ratio:float,
                    line_break_free_space_ratio:float,
                    lines_left_aligned_threshold:float,
                    lines_right_aligned_threshold:float,
                    lines_center_aligned_threshold:float):
        ''' Set horizontal spacing based on lines layout and page bbox.
        
        * The general spacing is determined by paragraph alignment and indentation.
        * The detailed spacing of block lines is determined by tab stops.

        Multiple alignment modes may exist in block (due to improper organized lines
        from ``PyMuPDF``), e.g. some lines align left, and others right. In this case,
        **LEFT** alignment is set, and use ``TAB`` to position each line.
        '''
        # NOTE: in PyMuPDF CS, horizontal text direction is same with positive x-axis,
        # while vertical text is on the contrary, so use f = -1 here
        idx0, idx1, f = (0, 2, 1.0) if self.is_horizontal_text else (3, 1, -1.0)
        
        # text alignment
        self.alignment = self._parse_alignment(bbox, 
            (idx0, idx1, f),
            line_separate_threshold,
            lines_left_aligned_threshold,
            lines_right_aligned_threshold,
            lines_center_aligned_threshold)

        # if still can't decide, set LEFT by default and ensure position by TAB stops        
        if self.alignment == TextAlignment.NONE:
            self.alignment = TextAlignment.LEFT
            self.lines.parse_tab_stop(line_separate_threshold)
        
        # adjust left/right indentation:
        # - set single side indentation if single line
        # - add minor space if multi-lines
        row_count = self.row_count
        if row_count==1 and self.alignment == TextAlignment.LEFT:
            self.right_space = 0

        elif row_count==1 and self.alignment == TextAlignment.RIGHT:
            self.left_space = 0
        
        elif row_count==1 and self.alignment == TextAlignment.CENTER:
            self.left_space = 0
            self.right_space = 0
        
        # parse line break
        self.lines.parse_line_break(bbox, 
            line_break_width_ratio, 
            line_break_free_space_ratio)


    def parse_relative_line_spacing(self):
        '''Calculate relative line spacing, e.g. `spacing = 1.02`.  Relative line spacing is based on standard 
        single line height, which is font-related. 

        .. note::
            The line spacing could be updated automatically when changing the font size, while the layout might
            be broken in exact spacing mode, e.g. overlapping of lines.
        '''
        # return default line spacing if any images exists
        for line in self.lines:
            if list(span for span in line.spans if isinstance(span, ImageSpan)):
                self.line_space = constants.DEFAULT_LINE_SPACING
                return

        # otherwise, calculate average line spacing
        idx = 1 if self.is_horizontal_text else 0
        block_height = self.bbox[idx+2]-self.bbox[idx]
        
        # An approximate expression: 
        # standard_line_height * relative_line_spacing = block_height
        rows = self.lines.group_by_physical_rows()        
        fun_max_line_height = lambda line: max(span.line_height for span in line.spans)
        fun_max_row_height = lambda row: max(fun_max_line_height(line) for line in row)
        standard_height = sum(fun_max_row_height(row) for row in rows)
        line_space = block_height/standard_height

        # overlap may exist when multi-rows, so set minimum spacing  -> default spacing
        if len(rows)>1: line_space = max(line_space, constants.DEFAULT_LINE_SPACING)
        self.line_space = line_space


    def parse_exact_line_spacing(self):
        '''Calculate exact line spacing, e.g. `spacing = Pt(12)`. 

        The layout of pdf text block: line-space-line-space-line, excepting space before first line, 
        i.e. space-line-space-line, when creating paragraph in docx. So, an average line height is 
        ``space+line``. Then, the height of first line can be adjusted by updating paragraph before-spacing.

        .. note::
            Compared with the relative spacing mode, it has a more precise layout, but less flexible editing
            ability, especially changing the font size.
        '''

        # check text direction
        idx = 1 if self.is_horizontal_text else 0       

        bbox = self.lines[0].bbox   # first line
        first_line_height = bbox[idx+2] - bbox[idx]
        block_height = self.bbox[idx+2]-self.bbox[idx]
        
        # average line spacing
        count = self.row_count # count of rows
        if count > 1:
            line_space = (block_height-first_line_height)/(count-1)
        else:
            line_space = block_height        
        self.line_space = line_space

        # since the line height setting in docx may affect the original bbox in pdf, 
        # it's necessary to update the before spacing:
        # taking bottom left corner of first line as the reference point                
        self.before_space += first_line_height - line_space

        # if before spacing is negative, set to zero and adjust calculated line spacing accordingly
        if self.before_space < 0:
            self.line_space += self.before_space / count
            self.before_space = 0.0


    def make_docx(self, p):
        '''Create paragraph for a text block.

        Refer to ``python-docx`` doc for details on text format:

        * https://python-docx.readthedocs.io/en/latest/user/text.html
        * https://python-docx.readthedocs.io/en/latest/api/enum/WdAlignParagraph.html#wdparagraphalignment
        
        Args:
            p (Paragraph): ``python-docx`` paragraph instance.

        .. note::
            The left position of paragraph is set by paragraph indent, rather than ``TAB`` stop.
        '''
        pf = docx.reset_paragraph_format(p)

        # ------------------------------------
        # vertical spacing
        # ------------------------------------
        before_spacing = max(round(self.before_space, 1), 0.0)
        after_spacing = max(round(self.after_space, 1), 0.0)
        pf.space_before = Pt(before_spacing)
        pf.space_after = Pt(after_spacing)        

        # line spacing
        if self.line_space_type==0: # exact line spacing
            pf.line_spacing = Pt(round(self.line_space, 1))
        else: # relative line spacing
            pf.line_spacing = round(self.line_space, 2)

        # ------------------------------------
        # horizontal alignment
        # ------------------------------------
        # (1) set paragraph indentation
        # NOTE: different left spacing setting in case first line indent and hanging
        left_space  = self.left_space        
        if self.first_line_space<0: # in case hanging
            left_space -= self.first_line_space           
        
        pf.left_indent  = Pt(left_space)
        pf.right_indent  = Pt(self.right_space)
        pf.first_line_indent = Pt(self.first_line_space)

        # (2) set alignment mode and adjust indentation:
        # round indention on the opposite side to lower bound (inches), so it saves more space to 
        # avoid unexpected line break
        if self.alignment==TextAlignment.LEFT:
            pf.alignment = WD_ALIGN_PARAGRAPH.LEFT            
            # set tab stops to ensure line position
            for pos in self.tab_stops:
                pf.tab_stops.add_tab_stop(Pt(self.left_space + pos))
            
            # adjust right indent
            d = lower_round(self.right_space/constants.ITP, 1)
            pf.right_indent = Inches(d)

        elif self.alignment==TextAlignment.RIGHT:
            pf.alignment = WD_ALIGN_PARAGRAPH.RIGHT
            
            # adjust left indent
            d = lower_round(left_space/constants.ITP, 1)
            pf.left_indent = Inches(d)

        elif self.alignment==TextAlignment.CENTER:
            pf.alignment = WD_ALIGN_PARAGRAPH.CENTER

            # adjust both left and right indent
            d = lower_round(left_space/constants.ITP, 1)
            pf.left_indent = Inches(d)
            d = lower_round(self.right_space/constants.ITP, 1)
            pf.right_indent = Inches(d)

        else:
            pf.alignment = WD_ALIGN_PARAGRAPH.JUSTIFY

        # ------------------------------------
        # add lines
        # ------------------------------------
        for line in self.lines: line.make_docx(p)

        return p



    def _parse_alignment(self, bbox,
                    text_direction_param:tuple, 
                    line_separate_threshold:float,
                    lines_left_aligned_threshold:float,
                    lines_right_aligned_threshold:float,
                    lines_center_aligned_threshold:float):
        '''Detect text alignment mode based on layout of internal lines. It can't decide when only
        one line, in such case, the alignment mode is determined by externally check.
        
        Args:
            text_direction_param (tuple): ``(x0_index, x1_index, direction_factor)``, 
                e.g. ``(0, 2, 1)`` for horizontal text, while ``(3, 1, -1)`` for vertical text.
        '''
        # indexes based on text direction
        idx0, idx1, f = text_direction_param

        # check external position: text block to parent layout
        d_left   = round((self.bbox[idx0]-bbox[idx0])*f, 1) # left margin
        d_right  = round((bbox[idx1]-self.bbox[idx1])*f, 1) # right margin
        d_center = round((d_left-d_right)/2.0, 1)           # center margin
        d_left   = max(d_left, 0.0)
        d_right  = max(d_right, 0.0)
        W = abs(bbox[idx1]-bbox[idx0]) # bbox width

        # NOTE: set horizontal space
        self.left_space  = d_left
        self.right_space = d_right

        # --------------------------------------------------------------------------
        # First priority: 
        # significant distance exists in any two adjacent lines -> set NONE temporarily. 
        # Assign left-align to it later and ensure exact position of each line by TAB stop. 
        # --------------------------------------------------------------------------
        rows = self.lines.group_by_physical_rows() # lines in each physical row
        for row in rows:
            if len(row)==1: continue
            dis = [(row[i].bbox[idx0]-row[i-1].bbox[idx1])*f>=line_separate_threshold \
                        for i in range(1, len(row))]
            if any(dis):
                return TextAlignment.NONE

        # --------------------------------------------------------------------------
        # determined by position to external bbox if one line only
        # the priority: center > left > right
        # --------------------------------------------------------------------------
        # |    ============    | -> center
        # |   ================ | -> left
        # |        =========== | -> right
        def external_alignment():         
            if abs(d_center) < lines_center_aligned_threshold: 
                return TextAlignment.CENTER
            elif d_left <= 0.25*W:
                return TextAlignment.LEFT
            else:
                return TextAlignment.RIGHT
        
        if len(rows) == 1: return external_alignment()

        # --------------------------------------------------------------------------
        # Check alignment of internal lines:
        # When count of lines >= 3:
        # - left-alignment based on lines excepting the first line
        # - right-alignment based on lines excepting the last line
        # the exact position of first line is considered by first-line-indent
        # =========     =======   =========  | =========   =========
        # =========   =========     =======  |   =======   =======
        # =========   =========     =======  |
        # ======      =========     =====    |
        # --------------------------------------------------------------------------
        X0 = [lines[0].bbox[idx0]  for lines in rows]
        X1 = [lines[-1].bbox[idx1] for lines in rows]
        X  = [(x0+x1)/2.0 for (x0, x1) in zip(X0, X1)]

        if len(rows) >= 3: X0, X1 = X0[1:], X1[0:-1]
        left_aligned   = abs(max(X0)-min(X0))<=lines_left_aligned_threshold
        right_aligned  = abs(max(X1)-min(X1))<=lines_right_aligned_threshold
        center_aligned = abs(max(X)-min(X))<=lines_center_aligned_threshold # coarse margin for center alignment

        if left_aligned and right_aligned:
            # need further external check if two lines only
            alignment = TextAlignment.JUSTIFY if len(rows)>=3 else external_alignment()

        elif center_aligned:
            alignment = TextAlignment.CENTER

        elif left_aligned:            
            alignment = TextAlignment.LEFT

        elif right_aligned:
            # change right alignment to left if two lines only
            alignment = TextAlignment.RIGHT if len(rows)>=3 else TextAlignment.LEFT

        else:
            alignment = TextAlignment.NONE
        
        # set first line space in case left/justify
        if alignment==TextAlignment.LEFT or alignment==TextAlignment.JUSTIFY:
            self.first_line_space = rows[0][0].bbox[idx0] - rows[1][0].bbox[idx0]
        
        return alignment


# -*- coding: utf-8 -*-

'''A group of Line objects.
'''


import string
from .Line import Line
from .TextSpan import TextSpan
from ..image.ImageSpan import ImageSpan
from ..common.Collection import ElementCollection
from ..common.share import TextAlignment
from ..common import constants


class Lines(ElementCollection):
    '''Collection of text lines.'''

    @property
    def unique_parent(self):
        '''Whether all contained lines have same parent.'''
        if not bool(self): return False

        first_line = self._instances[0]
        return all(line.same_source_parent(first_line) for line in self._instances)


    def restore(self, raws:list):
        '''Construct lines from raw dicts list.'''
        for raw in raws:
            line = Line(raw)
            self.append(line)
        return self


    @property
    def image_spans(self):
        '''Get all ImageSpan instances.'''
        spans = []
        for line in self._instances:
            spans.extend(line.image_spans)
        return spans

    
    def split_vertically_by_text(self, line_break_free_space_ratio:float, new_paragraph_free_space_ratio:float):
        '''Split lines into separate paragraph by checking text. The parent text block consists of 
        lines with similar line spacing, while lines in other paragraph might be counted when the
        paragraph spacing is relatively small. So, it's necessary to split those lines by checking
        the text contents.

        .. note::
            Considered only normal reading direction, from left to right, from top
            to bottom.
        '''
        rows = self.group_by_physical_rows()

        # skip if only one row
        num = len(rows)
        if num==1: return rows

        # standard row width with first row excluded, considering potential indentation of fist line
        W = max(row[-1].bbox[2]-row[0].bbox[0] for row in rows[1:])
        H = sum(row[0].bbox[3]-row[0].bbox[1] for row in rows) / num

        # check row by row
        res = []
        lines = Lines()
        punc = tuple(constants.SENTENCE_END_PUNC)
        start_of_para = end_of_para = False # start/end of paragraph
        start_of_sen = end_of_sen = False   # start/end of sentence
        for row in rows:
            end_of_sen = row[-1].text.strip().endswith(punc)
            w =  row[-1].bbox[2]-row[0].bbox[0]

            # end of a sentense and free space at the end -> end of paragraph
            if end_of_sen and w/W <= 1.0-line_break_free_space_ratio:
                end_of_para = True

            # start of sentence and free space at the start -> start of paragraph
            elif start_of_sen and (W-w)/H >= new_paragraph_free_space_ratio:
                start_of_para = True

            # take action
            if end_of_para:
                lines.extend(row)
                res.append(lines)
                lines = Lines()
            elif start_of_para:
                res.append(lines)
                lines = Lines()
                lines.extend(row)
            else:
                lines.extend(row)

            # for next round
            start_of_sen = end_of_sen
            start_of_para = end_of_para = False
        
        # close the action
        if lines: res.append(lines)

        return res


    def adjust_last_word(self, delete_end_line_hyphen:bool):
        '''Adjust word at the end of line:
        # - it might miss blank between words from adjacent lines
        # - it's optional to delete hyphen since it might not at the the end 
           of line after conversion
        '''
        punc_ex_hyphen = ''.join(c for c in string.punctuation if c!='-')
        def is_end_of_english_word(c):
            return c.isalnum() or (c and c in punc_ex_hyphen)
        
        for i, line in enumerate(self._instances[:-1]):
            # last char in this line
            end_span = line.spans[-1]
            if not isinstance(end_span, TextSpan): continue
            end_chars = end_span.chars
            if not end_chars: continue 
            end_char = end_chars[-1]

            # first char in next line
            start_span = self._instances[i+1].spans[0]
            if not isinstance(start_span, TextSpan): continue
            start_chars = start_span.chars
            if not start_chars: continue 
            next_start_char = start_chars[0]            

            # delete hyphen if next line starts with lower case letter
            if delete_end_line_hyphen and \
                end_char.c.endswith('-') and next_start_char.c.islower(): 
                end_char.c = '' # delete hyphen in a tricky way


            # add a space if both the last char and the first char in next line are alphabet,  
            # number, or English punctuation (excepting hyphen)
            if is_end_of_english_word(end_char.c) and is_end_of_english_word(next_start_char.c):
                end_char.c += ' ' # add blank in a tricky way


    def parse_text_format(self, shape):
        '''Parse text format with style represented by rectangle shape.
        
        Args:
            shape (Shape): Potential style shape applied on blocks.
        
        Returns:
            bool: Whether a valid text style.
        '''
        flag = False

        for line in self._instances:
            # any intersection in this line?
            expanded_bbox = line.get_expand_bbox(constants.MAJOR_DIST)
            if not shape.bbox.intersects(expanded_bbox): 
                if shape.bbox.y1 < line.bbox.y0: break # lines must be sorted in advance
                continue

            # yes, then try to split the spans in this line
            split_spans = []
            for span in line.spans: 
                # include image span directly
                if isinstance(span, ImageSpan): split_spans.append(span)                   

                # split text span with the format rectangle: span-intersection-span
                else:
                    spans = span.split(shape, line.is_horizontal_text)
                    split_spans.extend(spans)
                    flag = True
                                            
            # update line spans                
            line.spans.reset(split_spans)

        return flag


    def parse_line_break(self, bbox, 
                line_break_width_ratio:float, 
                line_break_free_space_ratio:float):
        '''Whether hard break each line. 

        Args:
            bbox (Rect): bbox of parent layout, e.g. page or cell.
            line_break_width_ratio (float): user defined threshold, break line if smaller than this value.
            line_break_free_space_ratio (float): user defined threshold, break line if exceeds this value.

        Hard line break helps ensure paragraph structure, but pdf-based layout calculation may
        change in docx due to different rendering mechanism like font, spacing. For instance, when
        one paragraph row can't accommodate a Line, the hard break leads to an unnecessary empty row.
        Since we can't 100% ensure a same structure, it's better to focus on the content - add line
        break only when it's necessary to, e.g. short lines.
        '''

        block = self.parent        
        idx0, idx1 = (0, 2) if block.is_horizontal_text else (3, 1)
        block_width = abs(block.bbox[idx1]-block.bbox[idx0])
        layout_width = bbox[idx1] - bbox[idx0]

        # hard break if exceed the width ratio
        line_break = block_width/layout_width <= line_break_width_ratio

        # check by each physical row
        rows = self.group_by_physical_rows()
        for lines in rows:
            for line in lines: line.line_break = 0

            # check the end line depending on text alignment
            if block.alignment == TextAlignment.RIGHT:
                end_line = lines[0]
                free_space = abs(block.bbox[idx0]-end_line.bbox[idx0])
            else:
                end_line = lines[-1]
                free_space = abs(block.bbox[idx1]-end_line.bbox[idx1])
            
            if block.alignment == TextAlignment.CENTER: free_space *= 2 # two side space
            
            # break line if 
            # - width ratio lower than the threshold; or 
            # - free space exceeds the threshold
            if line_break or free_space/block_width > line_break_free_space_ratio:
                end_line.line_break = 1
        
        # no break for last row
        for line in rows[-1]: line.line_break = 0


    def parse_tab_stop(self, line_separate_threshold:float):
        '''Calculate tab stops for parent block and whether add TAB stop before each line. 

        Args:
            line_separate_threshold (float): Don't need a tab stop if the line gap less than this value.
        '''
        # set all tab stop positions for parent block
        # Note these values are relative to the left boundary of parent block
        block = self.parent        
        idx0, idx1 = (0, 2) if block.is_horizontal_text else (3, 1)
        fun = lambda line: round(abs(line.bbox[idx0]-block.bbox[idx0]), 1)
        all_pos = set(map(fun, self._instances))
        tab_stops = list(filter(lambda pos: pos>=constants.MINOR_DIST, all_pos))
        tab_stops.sort() # sort in order
        block.tab_stops = tab_stops

        # no tab stop need
        if not block.tab_stops: return

        def tab_position(pos): # tab stop index of given position
            # 0   T0  <pos>  T1    T2
            i = 0
            pos -= block.bbox[idx0]
            for t in tab_stops:
                if pos<t: break
                i += 1
            return i

        # otherwise, set tab stop option for each line
        # Note: it might need more than one tab stops
        # https://github.com/dothinking/pdf2docx/issues/157
        ref = block.bbox[idx0]
        for i, line in enumerate(self._instances):
            # left indentation implemented with tab
            distance = line.bbox[idx0] - ref
            if distance>line_separate_threshold:
                line.tab_stop = tab_position(line.bbox[idx0])-tab_position(ref)

            # update stop reference position
            if line==self._instances[-1]: break
            ref = line.bbox[idx1] if line.in_same_row(self._instances[i+1]) else block.bbox[idx0]


# -*- coding: utf-8 -*-

'''Base class for text/image/table blocks.
'''

from .share import BlockType, TextAlignment
from .Element import Element


class Block(Element):
    '''Base class for text/image/table blocks.

    Attributes:
        raw (dict): initialize object from raw properties.
        parent (optional): parent object that this block belongs to.
    '''
    def __init__(self, raw:dict=None, parent=None):        
        self._type = BlockType.UNDEFINED

        # horizontal spacing
        if raw is None: raw = {}
        self.alignment = self._get_alignment(raw.get('alignment', 0))
        self.left_space = raw.get('left_space', 0.0)
        self.right_space = raw.get('right_space', 0.0)
        self.first_line_space = raw.get('first_line_space', 0.0)

        # RELATIVE position of tab stops
        self.tab_stops = raw.get('tab_stops', []) 

        # vertical spacing
        self.before_space = raw.get('before_space', 0.0)
        self.after_space = raw.get('after_space', 0.0)        
        self.line_space = raw.get('line_space', 0.0)
        self.line_space_type = raw.get('line_space_type', 1) # 0-exactly, 1-relatively

        super().__init__(raw, parent)


    @property
    def is_text_block(self):
        '''Whether test block.'''        
        return self._type==BlockType.TEXT    
    
    @property
    def is_inline_image_block(self):
        '''Whether inline image block.'''
        return self._type==BlockType.IMAGE
    
    @property
    def is_float_image_block(self):
        '''Whether float image block.'''
        return self._type==BlockType.FLOAT_IMAGE
    
    @property
    def is_image_block(self):
        '''Whether inline or float image block.'''
        return self.is_inline_image_block or self.is_float_image_block

    @property
    def is_text_image_block(self):
        '''Whether text block or inline image block.'''
        return self.is_text_block or self.is_inline_image_block

    @property
    def is_lattice_table_block(self):
        '''Whether lattice table (explicit table borders) block.'''
        return self._type==BlockType.LATTICE_TABLE

    @property
    def is_stream_table_block(self):
        '''Whether stream table (implied by table content) block.'''
        return self._type==BlockType.STREAM_TABLE

    @property
    def is_table_block(self):
        '''Whether table (lattice or stream) block.'''
        return self.is_lattice_table_block or self.is_stream_table_block

    def set_text_block(self):
        '''Set block type.'''
        self._type = BlockType.TEXT

    def set_inline_image_block(self):
        '''Set block type.'''
        self._type = BlockType.IMAGE

    def set_float_image_block(self):
        '''Set block type.'''
        self._type = BlockType.FLOAT_IMAGE

    def set_lattice_table_block(self):
        '''Set block type.'''
        self._type = BlockType.LATTICE_TABLE

    def set_stream_table_block(self):
        '''Set block type.'''
        self._type = BlockType.STREAM_TABLE

    def _get_alignment(self, mode:int):
        for t in TextAlignment:
            if t.value==mode:
                return t
        return TextAlignment.LEFT

    def parse_horizontal_spacing(self, bbox, *args):
        """Set left alignment, and calculate left space. 
        
        Override by :obj:`pdf2docx.text.TextBlock`.

        Args:
            bbox (fitz.rect): boundary box of this block.
        """
        # NOTE: in PyMuPDF CS, horizontal text direction is same with positive x-axis,
        # while vertical text is on the contrary, so use f = -1 here
        idx, f = (0, 1.0) if self.is_horizontal_text else (3, -1.0)
        self.alignment = TextAlignment.LEFT
        self.left_space = (self.bbox[idx] - bbox[idx]) * f
       

    def store(self):
        '''Store attributes in json format.'''
        res = super().store()
        res.update({
            'type'             : self._type.value,
            'alignment'        : self.alignment.value,
            'left_space'       : self.left_space,
            'right_space'      : self.right_space,
            'first_line_space' : self.first_line_space,
            'before_space'     : self.before_space,
            'after_space'      : self.after_space,
            'line_space'       : self.line_space,
            'line_space_type'  : self.line_space_type,
            'tab_stops'        : self.tab_stops
            })
        return res


    def make_docx(self, *args, **kwargs):
        """Create associated docx element.

        Raises:
            NotImplementedError
        """
        raise NotImplementedError


# -*- coding: utf-8 -*-

'''Table block object parsed from raw image and text blocks.

Data Structure::

    {
        'type': int
        'bbox': (x0, y0, x1, y1),
        'rows': [
            {
                "bbox": (x0, y0, x1, y1),
                "height": float,
                "cells": [
                    {
                        'bbox': (x0, y0, x1, y1),
                        'border_color': (sRGB,,,), # top, right, bottom, left
                        'bg_color': sRGB,
                        'border_width': (,,,),
                        'merged_cells': (x,y), # this is the bottom-right cell of merged region: x rows, y cols
                        'blocks': [ {text blocks} ]
                    }, # end of cell
                    {},
                    None, # merged cell
                    ...
                ]
            }, # end of row
            {...} # more rows
        ] # end of row
    }
'''


from .Row import Row
from .Rows import Rows
from ..common.Block import Block
from ..common import docx


class TableBlock(Block):
    '''Table block.'''
    def __init__(self, raw:dict=None):
        if raw is None: raw = {}
        super().__init__(raw)

        # collect rows
        self._rows = Rows(parent=self).restore(raw.get('rows', []))

        # lattice table by default
        self.set_lattice_table_block()

    def __getitem__(self, idx):
        try:
            row = self._rows[idx]
        except IndexError:
            msg = f'Row index {idx} out of range'
            raise IndexError(msg)
        else:
            return row

    def __iter__(self):
        return (row for row in self._rows)

    def __len__(self):
        return len(self._rows)

    @property
    def num_rows(self):
        '''Count of rows.'''
        return len(self._rows)

    @property
    def num_cols(self):
        '''Count of columns.'''
        return len(self._rows[0]) if self.num_rows else 0

    @property
    def text(self):
        '''Get text contained in each cell.

        Returns:
            list: 2D-list with each element representing text in cell.
        '''
        return [ [cell.text for cell in row] for row in self._rows ]

    @property
    def outer_bbox(self):
        '''Outer bbox with border considered.'''
        x0, y0, x1, y1 = self.bbox
        w0_top, w0_right, w0_bottom, w0_left = self[0][0].border_width
        w1_top, w1_right, w1_bottom, w1_left = self[-1][-1].border_width
        return (x0-w0_left/2.0, y0-w0_top/2.0, x1+w1_right/2.0, y1+w1_bottom/2.0)


    def append(self, row:Row):
        '''Append row to table and update bbox accordingly.

        Args:
            row (Row): Target row to add.
        '''
        self._rows.append(row)


    def store(self):
        res = super().store()
        res.update({
            'rows': self._rows.store()
        })
        return res


    def assign_blocks(self, blocks:list):
        '''Assign ``blocks`` to associated cell.

        Args:
            blocks (list): A list of text/table blocks.
        '''
        for row in self._rows:
            for cell in row:
                if not cell: continue
                cell.assign_blocks(blocks)


    def assign_shapes(self, shapes:list):
        '''Assign ``shapes`` to associated cell.

        Args:
            shapes (list): A list of Shape.
        '''
        for row in self._rows:
            for cell in row:
                if not cell: continue
                cell.assign_shapes(shapes)


    def parse(self, **settings):
        '''Parse layout under cell level.

        Args:
            settings (dict): Layout parsing parameters.
        '''
        for row in self._rows:
            for cell in row:
                if not cell: continue
                cell.parse(**settings)


    def plot(self, page):
        '''Plot table block, i.e. cell/line/span, for debug purpose.
        
        Args:
            page (fitz.Page): pdf page.
            content (bool): Plot text blocks contained in cells if True.
            style (bool): Plot cell style if True, e.g. border width, shading.
            color (bool): Plot border stroke color if ``style=False``.
        '''
        for row in self._rows:
            for cell in row:                
                if not cell: continue  # ignore merged cells   
                cell.plot(page)


    def make_docx(self, table):
        '''Create docx table.
        
        Args:
            table (Table): ``python-docx`` table instance.
        '''
        # set left indent
        docx.indent_table(table, self.left_space)

        # set format and contents row by row
        for idx_row in range(len(table.rows)):
            self._rows[idx_row].make_docx(table, idx_row)


# -*- coding: utf-8 -*-

'''Table Cell object.
'''

from docx.shared import Pt
from ..common.Element import Element
from ..layout.Layout import Layout
from ..common import docx


class Cell(Element, Layout):
    '''Cell object.'''
    def __init__(self, raw:dict=None):
        raw = raw or {}
        Element.__init__(self, raw=raw)
        self.bg_color     = raw.get('bg_color', None) # type: int
        self.border_color = raw.get('border_color', (0,0,0,0)) # type: tuple [int]
        self.border_width = raw.get('border_width', (0,0,0,0)) # type: tuple [float]
        self.merged_cells = raw.get('merged_cells', (1,1)) # type: tuple [int]

        # restore layout
        Layout.__init__(self, blocks=None, shapes=None) # init empty layout
        self.restore(raw)


    @property
    def text(self):
        '''Text contained in this cell.'''
        if not self: return None
        # NOTE: sub-table may exists in 
        return '\n'.join([block.text if block.is_text_block else '<NEST TABLE>'
                                 for block in self.blocks])


    @property
    def working_bbox(self):
        '''Inner bbox with border excluded.'''
        x0, y0, x1, y1 = self.bbox
        w_top, w_right, w_bottom, w_left = self.border_width
        bbox = (x0+w_left/2.0, y0+w_top/2.0, x1-w_right/2.0, y1-w_bottom/2.0)
        return Element().update_bbox(bbox).bbox # convert to fitz.Rect


    def store(self):
        if bool(self):
            res = super().store() # Element
            res.update({
                'bg_color': self.bg_color,
                'border_color': self.border_color,
                'border_width': self.border_width,
                'merged_cells': self.merged_cells
            })
            res.update(Layout.store(self))
            return res
        else:
            return None


    def plot(self, page):
        '''Plot cell and its sub-layout.'''        
        super().plot(page)
        self.blocks.plot(page)


    def make_docx(self, table, indexes):
        '''Set cell style and assign contents.
        
        Args:
            table (Table): ``python-docx`` table instance.
            indexes (tuple): Row and column indexes, ``(i, j)``.
        '''        
        # set cell style, e.g. border, shading, cell width
        self._set_style(table, indexes)
        
        # ignore merged cells
        if not bool(self):  return

        # merge cells
        n_row, n_col = self.merged_cells
        i, j = indexes
        docx_cell = table.cell(i, j)
        if n_row*n_col!=1:
            _cell = table.cell(i+n_row-1, j+n_col-1)
            docx_cell.merge(_cell)
        
        # ---------------------
        # cell width (cell height is set by row height)
        # ---------------------
        # experience: width of merged cells may change if not setting width for merged cells
        x0, y0, x1, y1 = self.bbox
        docx_cell.width = Pt(x1-x0)

        # insert contents
        # NOTE: there exists an empty paragraph already in each cell, which should be deleted first to
        # avoid unexpected layout. `docx_cell._element.clear_content()` works here.
        # But, docx requires at least one paragraph in each cell, otherwise resulting in a repair error. 
        if self.blocks:
            docx_cell._element.clear_content()
            self.blocks.make_docx(docx_cell)


    def _set_style(self, table, indexes):
        '''Set ``python-docx`` cell style, e.g. border, shading, width, row height, 
        based on cell block parsed from PDF.
        
        Args:
            table (Table): ``python-docx`` table object.
            indexes (tuple): ``(i, j)`` index of current cell in table.
        '''
        i, j = indexes
        docx_cell = table.cell(i, j)
        n_row, n_col = self.merged_cells

        # ---------------------
        # border style
        # ---------------------
        # NOTE: border width is specified in eighths of a point, with a minimum value of 
        # two (1/4 of a point) and a maximum value of 96 (twelve points)
        keys = ('top', 'end', 'bottom', 'start')
        kwargs = {}
        for k, w, c in zip(keys, self.border_width, self.border_color):
            # skip if width=0 -> will not show in docx
            if not w: continue

            hex_c = f'#{hex(c)[2:].zfill(6)}'
            kwargs[k] = {
                'sz': 8*w, 'val': 'single', 'color': hex_c.upper()
            }

        # merged cells are assumed to have same borders with the main cell        
        for m in range(i, i+n_row):
            for n in range(j, j+n_col):
                docx.set_cell_border(table.cell(m, n), **kwargs)        

        # ---------------------
        # cell bg-color
        # ---------------------
        if self.bg_color!=None:
            docx.set_cell_shading(docx_cell, self.bg_color)
        
        # ---------------------
        # clear cell margin
        # ---------------------
        # NOTE: the start position of a table is based on text in cell, rather than left border of table. 
        # They're almost aligned if left-margin of cell is zero.
        docx.set_cell_margins(docx_cell, start=0, end=0)

        # set vertical direction if contained text blocks are in vertical direction
        if self.blocks.is_vertical_text:
            docx.set_vertical_cell_direction(docx_cell)


# -*- coding: utf-8 -*-

'''Char object based on PDF raw dict extracted with ``PyMuPDF``.

Data structure refer to this `link <https://pymupdf.readthedocs.io/en/latest/textpage.html>`_::

    {
        'bbox'  : (x0, y0, x1, y1), 
        'c'     : str, 
        'origin': (x,y)
    }
'''


from ..common.constants import INVALID_CHARS
from ..common.Element import Element
from ..shape.Shape import Shape


class Char(Element):
    '''Object representing a character.'''
    def __init__(self, raw:dict=None):
        if raw is None: raw = {}

        # Note to filter control character avoiding error when making docx, #126
        c = raw.get('c', '')
        if c in INVALID_CHARS: c = ''
        self.c = c
        self.origin = raw.get('origin', None)

        super().__init__(raw) # NOTE: ignore parent element for Char instance


    def contained_in_rect(self, rect:Shape, horizontal:bool=True):
        """Detect whether it locates in a rect.

        Args:
            rect (Shape): Target rect to check.
            horizontal (bool, optional): Text direction is horizontal if True. Defaults to True.

        Returns:
            bool: Whether a Char locates in target rect.
        
        .. note::
            It's considered as contained in the target rect if the intersection is larger than 
            half of the char bbox.
        """ 
        # char in rect?
        if self.bbox in rect.bbox:
            return True

        # intersection? 
        else:
            intsec = self.bbox & rect.bbox # width=0 if invalid intersection
            if horizontal:
                return intsec.width > 0.5*self.bbox.width
            else:
                return intsec.height > 0.5*self.bbox.height


    def store(self):
        res = super().store()
        res.update({
            'c': self.c,
            'origin': self.origin
        })

        return res


# -*- coding: utf-8 -*-
import json
import logging
import os
from multiprocessing import Pool, cpu_count
from time import perf_counter
from typing import AnyStr, IO, Union

import fitz
from docx import Document

from .page.Page import Page
from .page.Pages import Pages

# check PyMuPDF>=1.19.x
if list(map(int, fitz.VersionBind.split("."))) < [1, 19, 0]:
    raise SystemExit("PyMuPDF>=1.19.0 is required for pdf2word.")

# logging
logging.basicConfig(
    level=logging.INFO, 
    format="[%(levelname)s] %(message)s")


class Converter:
    '''The ``PDF`` to ``docx`` converter.
    
    * Read PDF file with ``PyMuPDF`` to get raw layout data page by page, including text,
      image, drawing and its properties, e.g. boundary box, font, size, image width, height.
    * Analyze layout in document level, e.g. page header, footer and margin.
    * Parse page layout to docx structure, e.g. paragraph and its properties like indentation, 
      spacing, text alignment; table and its properties like border, shading, merging. 
    * Finally, generate docx with ``python-docx``.
    '''

    def __init__(
        self, pdf_file: str = None, password: str = None, stream: bytes = None
    ):
        '''Initialize fitz object with given pdf file path.

        Args:
            pdf_file (str): pdf file path.
            stream   (bytes): pdf file in memory.
            password (str): Password for encrypted pdf. Default to None if not encrypted.
        '''
        # fitz object
        self.filename_pdf = pdf_file
        self.password = str(password or "")

        if not pdf_file and not stream:
            raise ValueError("Either pdf_file or stream must be given.")

        if stream:
            self._fitz_doc = fitz.Document(stream=stream)

        else:
            self._fitz_doc = fitz.Document(pdf_file)

        # initialize empty pages container
        self._pages = Pages()


    @property
    def fitz_doc(self): return self._fitz_doc    

    @property
    def pages(self): return self._pages


    def close(self): self._fitz_doc.close()


    @property
    def default_settings(self):
        '''Default parsing parameters.'''
        return {
            'debug'                          : False,  # plot layout if True
            'ocr'                            : 0,      # ocr status: 0 - no ocr; 1 - to do ocr; 2 - ocr-ed pdf
            'ignore_page_error'              : True,   # not break the conversion process due to failure of a certain page if True
            'multi_processing'               : False,  # convert pages with multi-processing if True
            'cpu_count'                      : 0,      # working cpu count when convert pages with multi-processing
            'min_section_height'             : 20.0,   # The minimum height of a valid section.
            'connected_border_tolerance'     : 0.5,    # two borders are intersected if the gap lower than this value
            'max_border_width'               : 6.0,    # max border width
            'min_border_clearance'           : 2.0,    # the minimum allowable clearance of two borders
            'float_image_ignorable_gap'      : 5.0,    # float image if the intersection exceeds this value
            'page_margin_factor_top'         : 0.5,    # [0,1] reduce top margin by factor
            'page_margin_factor_bottom'      : 0.5,    # [0,1] reduce bottom margin by factor
            'shape_min_dimension'            : 2.0,    # ignore shape if both width and height is lower than this value
            'max_line_spacing_ratio'         : 1.5,    # maximum line spacing ratio: line spacing / line height
            'line_overlap_threshold'         : 0.9,    # [0,1] delete line if the intersection to other lines exceeds this value
            'line_break_width_ratio'         : 0.5,    # break line if the ratio of line width to entire layout bbox is lower than this value
            'line_break_free_space_ratio'    : 0.1,    # break line if the ratio of free space to entire line exceeds this value
            'line_separate_threshold'        : 5.0,    # two separate lines if the x-distance exceeds this value
            'new_paragraph_free_space_ratio' : 0.85,   # new paragraph if the ratio of free space to line height exceeds this value
            'lines_left_aligned_threshold'   : 1.0,    # left aligned if d_x0 of two lines is lower than this value (Pt)
            'lines_right_aligned_threshold'  : 1.0,    # right aligned if d_x1 of two lines is lower than this value (Pt)
            'lines_center_aligned_threshold' : 2.0,    # center aligned if delta center of two lines is lower than this value
            'clip_image_res_ratio'           : 4.0,    # resolution ratio (to 72dpi) when clipping page image
            'min_svg_gap_dx'                 : 15.0,   # merge adjacent vector graphics if the horizontal gap is less than this value
            'min_svg_gap_dy'                 : 2.0,    # merge adjacent vector graphics if the vertical gap is less than this value
            'min_svg_w'                      : 2.0,    # ignore vector graphics if the bbox width is less than this value
            'min_svg_h'                      : 2.0,    # ignore vector graphics if the bbox height is less than this value
            'extract_stream_table'           : False,  # don't consider stream table when extracting tables
            'parse_lattice_table'            : True,   # whether parse lattice table or not; may destroy the layout if set False
            'parse_stream_table'             : True,   # whether parse stream table or not; may destroy the layout if set False
            'delete_end_line_hyphen'         : False   # delete hyphen at the end of a line
        }

    # -----------------------------------------------------------------------
    # Parsing process: load -> analyze document -> parse pages -> make docx
    # -----------------------------------------------------------------------

    def parse(self, start:int=0, end:int=None, pages:list=None, **kwargs):
        '''Parse pages in three steps:
        * open PDF file with ``PyMuPDF``
        * analyze whole document, e.g. page section, header/footer and margin
        * parse specified pages, e.g. paragraph, image and table

        Args:
            start (int, optional): First page to process. Defaults to 0, the first page.
            end (int, optional): Last page to process. Defaults to None, the last page.
            pages (list, optional): Range of page indexes to parse. Defaults to None.
            kwargs (dict, optional): Configuration parameters. 
        '''
        return self.load_pages(start, end, pages) \
            .parse_document(**kwargs) \
            .parse_pages(**kwargs)


    def load_pages(self, start:int=0, end:int=None, pages:list=None):
        '''Step 1 of converting process: open PDF file with ``PyMuPDF``, 
        especially for password encrypted file.
        
        Args:
            start (int, optional): First page to process. Defaults to 0, the first page.
            end (int, optional): Last page to process. Defaults to None, the last page.
            pages (list, optional): Range of page indexes to parse. Defaults to None.
        '''
        logging.info(self._color_output('[1/4] Opening document...'))

        # encrypted pdf ?
        if self._fitz_doc.needs_pass:
            if not self.password:
                raise ConversionException(f'Require password for {self.filename_pdf}.')

            elif not self._fitz_doc.authenticate(self.password):
                raise ConversionException('Incorrect password.')

        # initialize empty pages
        num = len(self._fitz_doc)
        self._pages.reset([Page(id=i, skip_parsing=True) for i in range(num)])

        # set pages to parse
        page_indexes = self._page_indexes(start, end, pages, num)
        for i in page_indexes:
            self._pages[i].skip_parsing = False

        return self
    

    def parse_document(self, **kwargs):
        '''Step 2 of converting process: analyze whole document, e.g. page section,
        header/footer and margin.'''
        logging.info(self._color_output('[2/4] Analyzing document...'))
        
        self._pages.parse(self.fitz_doc, **kwargs)
        return self

    
    def parse_pages(self, **kwargs):
        '''Step 3 of converting process: parse pages, e.g. paragraph, image and table.'''
        logging.info(self._color_output('[3/4] Parsing pages...'))

        pages = [page for page in self._pages if not page.skip_parsing]
        num_pages = len(pages)
        for i, page in enumerate(pages, start=1):
            pid = page.id + 1
            logging.info('(%d/%d) Page %d', i, num_pages, pid)
            try:
                page.parse(**kwargs)
            except Exception as e:
                if not kwargs['debug'] and kwargs['ignore_page_error']:
                    logging.error('Ignore page %d due to parsing page error: %s', pid, e)
                else:
                    raise ConversionException(f'Error when parsing page {pid}: {e}')

        return self


    def make_docx(self, docx_filename=None, **kwargs):
        '''Step 4 of converting process: create docx file with converted pages.
        
        Args:
            docx_filename (str, file-like): docx file to write.
            kwargs (dict, optional): Configuration parameters. 
        '''
        logging.info(self._color_output('[4/4] Creating pages...'))

        # check parsed pages
        parsed_pages = list(filter(
            lambda page: page.finalized, self._pages
        ))
        if not parsed_pages:
            raise ConversionException('No parsed pages. Please parse page first.')

        if not docx_filename:
            raise ConversionException(
                "No docx file name. Please specify a docx file name or a file-like object to write."
            )

        # docx file to convert to
        if hasattr(docx_filename, "write"):
            filename = docx_filename

        else:
            filename = docx_filename or f'{self.filename_pdf[0:-len(".pdf")]}.docx' if self.filename_pdf else "output.docx"
            if os.path.exists(filename): os.remove(filename)

        # create page by page        
        docx_file = Document() 
        num_pages = len(parsed_pages)
        for i, page in enumerate(parsed_pages, start=1):
            if not page.finalized: continue # ignore unparsed pages
            pid = page.id + 1
            logging.info('(%d/%d) Page %d', i, num_pages, pid)
            try:
                page.make_docx(docx_file)
            except Exception as e:
                if not kwargs['debug'] and kwargs['ignore_page_error']:
                    logging.error('Ignore page %d due to making page error: %s', pid, e)
                else:
                    raise MakedocxException(f'Error when make page {pid}: {e}')

        # save docx
        docx_file.save(filename)


    # -----------------------------------------------------------------------
    # Store / restore parsed results
    # -----------------------------------------------------------------------

    def store(self):
        '''Store parsed pages in dict format.'''
        return {
            'filename': os.path.basename(self.filename_pdf),
            'page_cnt': len(self._pages), # count of all pages
            'pages'   : [page.store() for page in self._pages if page.finalized], # parsed pages only
        }


    def restore(self, data:dict):
        '''Restore pages from parsed results.'''
        # init empty pages if necessary
        if not self._pages:
            num = data.get('page_cnt', 100)
            self._pages.reset([Page(id=i, skip_parsing=True) for i in range(num)])
        
        # restore pages
        for raw_page in data.get('pages', []):
            idx = raw_page.get('id', -1)
            self._pages[idx].restore(raw_page)


    def serialize(self, filename:str):
        '''Write parsed pages to specified JSON file.'''
        with open(filename, 'w', encoding='utf-8') as f:
            f.write(json.dumps(self.store(), indent=4))
    

    def deserialize(self, filename:str):
        '''Load parsed pages from specified JSON file.'''
        with open(filename, 'r') as f:
            data = json.load(f)
        self.restore(data)


    # -----------------------------------------------------------------------
    # high level methods, e.g. convert, extract table
    # -----------------------------------------------------------------------

    def debug_page(self, i:int, docx_filename:str=None, debug_pdf:str=None, layout_file:str=None, **kwargs):
        '''Parse, create and plot single page for debug purpose.
        
        Args:
            i (int): Page index to convert.
            docx_filename (str): docx filename to write to.
            debug_pdf (str): New pdf file storing layout information. Default to add prefix ``debug_``.
            layout_file (str): New json file storing parsed layout data. Default to ``layout.json``.
        '''
        # include debug information
        # fitz object in debug mode: plot page layout
        # file path for this debug pdf: demo.pdf -> debug_demo.pdf
        path, filename = os.path.split(self.filename_pdf)
        if not debug_pdf: debug_pdf = os.path.join(path, f'debug_{filename}')
        if not layout_file: layout_file  = os.path.join(path, 'layout.json')
        kwargs.update({
            'debug'         : True,
            'debug_doc'     : fitz.Document(),
            'debug_filename': debug_pdf
        })

        # parse and create docx
        self.convert(docx_filename, pages=[i], **kwargs)
        
        # layout information for debugging
        self.serialize(layout_file)

    def convert(self, docx_filename: Union[str, IO[AnyStr]] = None, start: int = 0, end: int = None, pages: list = None,
                **kwargs):
        """Convert specified PDF pages to docx file.

        Args:
            docx_filename (str, file-like, optional): docx file to write. Defaults to None.
            start (int, optional): First page to process. Defaults to 0, the first page.
            end (int, optional): Last page to process. Defaults to None, the last page.
            pages (list, optional): Range of page indexes. Defaults to None.
            kwargs (dict, optional): Configuration parameters. Defaults to None.
        
        Refer to :py:meth:`~pdf2word.converter.Converter.default_settings` for detail of 
        configuration parameters.
        
        .. note::
            Change extension from ``pdf`` to ``docx`` if ``docx_file`` is None.
        
        .. note::
            * ``start`` and ``end`` is counted from zero if ``--zero_based_index=True`` (by default).
            * Start from the first page if ``start`` is omitted.
            * End with the last page if ``end`` is omitted.
        
        .. note::
            ``pages`` has a higher priority than ``start`` and ``end``. ``start`` and ``end`` works only
            if ``pages`` is omitted.

        .. note::
            Multi-processing works only for continuous pages specified by ``start`` and ``end`` only.
        """
        t0 = perf_counter()
        logging.info('Start to convert %s', self.filename_pdf)
        settings = self.default_settings
        settings.update(kwargs)

        # input check
        if pages and settings['multi_processing']:
            raise ConversionException('Multi-processing works for continuous pages '
                                    'specified by "start" and "end" only.')
        
        # convert page by page
        if settings['multi_processing']:
            self._convert_with_multi_processing(docx_filename, start, end, **settings)
        else:
            self.parse(start, end, pages, **settings).make_docx(docx_filename, **settings)

        logging.info('Terminated in %.2fs.', perf_counter()-t0)        


    def extract_tables(self, start:int=0, end:int=None, pages:list=None, **kwargs):
        '''Extract table contents from specified PDF pages.

        Args:
            start (int, optional): First page to process. Defaults to 0, the first page.
            end (int, optional): Last page to process. Defaults to None, the last page.
            pages (list, optional): Range of page indexes. Defaults to None.
            kwargs (dict, optional): Configuration parameters. Defaults to None.
        
        Returns:
            list: A list of parsed table content.
        '''
        # parsing pages first
        settings = self.default_settings
        settings.update(kwargs)
        self.parse(start, end, pages, **settings)

        # get parsed tables
        tables = []
        for page in self._pages:
            if page.finalized: tables.extend(page.extract_tables(**settings))

        return tables

    
    def _convert_with_multi_processing(self, docx_filename:str, start:int, end:int, **kwargs):
        '''Parse and create pages based on page indexes with multi-processing.

        Reference:

            https://pymupdf.readthedocs.io/en/latest/faq.html#multiprocessing
        '''
        # make vectors of arguments for the processes
        cpu = min(kwargs['cpu_count'], cpu_count()) if kwargs['cpu_count'] else cpu_count()        
        prefix = 'pages' # json file writing parsed pages per process
        vectors = [(i, cpu, start, end, self.filename_pdf, self.password, 
                            kwargs, f'{prefix}-{i}.json') for i in range(cpu)]

        # start parsing processes
        pool = Pool()
        pool.map(self._parse_pages_per_cpu, vectors, 1)
        
        # restore parsed page data
        for i in range(cpu):
            filename = f'{prefix}-{i}.json'
            if not os.path.exists(filename): continue            
            self.deserialize(filename)
            os.remove(filename)
        
        # create docx file
        self.make_docx(docx_filename, **kwargs)


    @staticmethod
    def _parse_pages_per_cpu(vector):
        '''Render a page range of a document.
        
        Args:
            vector (list): A list containing required parameters.
                * 0  : segment number for current process                
                * 1  : count of CPUs
                * 2,3: whole pages range to process
                * 4  : pdf filename
                * 5  : password for encrypted pdf
                * 6  : configuration parameters
                * 7  : json filename storing parsed results
        '''        
        # recreate the arguments
        idx, cpu, s, e, pdf_filename, password, kwargs, json_filename = vector

        # open pdf to get page count: all pages are marked to parse temporarily 
        # since don't know which pages to parse for this moment
        cv = Converter(pdf_filename, password)
        cv.load_pages()

        # the specified pages to process
        e = e or len(cv.fitz_doc)
        all_indexes = range(s, e)
        num_pages = len(all_indexes)

        # page segment processed by this cpu
        m = int(num_pages/cpu)
        n = num_pages % cpu
        seg_size = m + int(idx<n)
        seg_from = (m+1)*idx + min(n-idx, 0)
        seg_to = min(seg_from + seg_size, num_pages)
        page_indexes = [all_indexes[i] for i in range(seg_from, seg_to)]

        # now, mark the right pages
        for page in cv.pages: page.skip_parsing = True
        for i in page_indexes: 
            cv.pages[i].skip_parsing = False

        # parse pages and serialize data for further processing
        cv.parse_document(**kwargs) \
            .parse_pages(**kwargs) \
            .serialize(json_filename)
        cv.close()


    @staticmethod
    def _page_indexes(start, end, pages, pdf_len):
        '''Parsing arguments.'''
        if pages: 
            indexes = [int(x) for x in pages]
        else:
            end = end or pdf_len
            s = slice(int(start), int(end))
            indexes = range(pdf_len)[s]
        
        return indexes

    
    @staticmethod
    def _color_output(msg): return f'\033[1;36m{msg}\033[0m'


class ConversionException(Exception): 
    pass

class MakedocxException(ConversionException): 
    pass



# -*- coding: utf-8 -*-

'''Image span based on same raw data structure with Image block.
'''

from ..common import constants
from .Image import Image


class ImageSpan(Image):
    '''Image span.'''

    def intersects(self, rect):
        '''Create new ImageSpan object with image contained in given bbox.
        
        Args:
            rect (fitz.Rect): Target bbox.
        
        Returns:
            ImageSpan: A copy of itself if intersects with target; otherwise empty ImageSpan. 
        '''
        # add image span if most of of the image is contained in bbox
        if self.get_main_bbox(rect, constants.FACTOR_MAJOR):
            return self.copy()
        
        # otherwise, ignore image
        return ImageSpan()


'''Extract fonts properties from PDF.

Font properties like font name, size are covered in :py:class:`~pdf2docx.text.TextSpan`, 
but more generic properties are required further:

* Font family name. The font name extracted and set in ``TextSpan`` might not valid when 
  directly used in MS Word, e.g. "ArialMT" should be "Arial". So, we need to get font
  family name, which should be accepted by MS Word, based on the font file itself.

* Font line height ratio. As line height = font_size * line_height_ratio, it's used to 
  calculate relative line spacing. In general, 1.12 is an approximate value to this ratio,
  but it's in fact a font-related value, especially for CJK font. 

    * So, extract font metrics, e.g. ascender and descender, with third party library ``fontTools``
      in first priority. This can obtain an accurate line height ratio, but sometimes the 
      embedded font data might crash.
  
    * Then, we have to use the default properties, i.e. ascender and descender, extracted by
      ``PyMuPDF`` directly, but this value isn't so accurate.    
'''

import os
from io import BytesIO
from collections import namedtuple 
from fontTools.ttLib import TTFont
from ..common.Collection import BaseCollection
from ..common.constants import (CJK_CODEPAGE_BITS, CJK_UNICODE_RANGE_BITS, CJK_UNICODE_RANGES)


Font = namedtuple('Font', [ 'descriptor',     # font descriptor
                            'name',           # real font name
                            'line_height'])   # standard line height ratio


class Fonts(BaseCollection):
    '''Extracted fonts properties from PDF.'''

    def get(self, font_name:str):
        '''Get matched font by font name, or return None.'''
        target = self._to_descriptor(font_name)

        # 1st priority: check right the name
        for font in self:
            if target==font.descriptor: return font
        
        # 2nd priority: target name is contained in font name
        for font in self:
            if target in font.descriptor: return font

        # 3rd priority: target name contains font name
        for font in self:
            if font.descriptor in target: return font
        
        return None


    @classmethod
    def extract(cls, fitz_doc):
        '''Extract fonts from PDF and get properties.
        * Only embedded fonts (v.s. the base 14 fonts) can be extracted.
        * The extracted fonts may be invalid due to reason from PDF file itself.
        '''        
        # get unique font references
        xrefs = set()
        for page in fitz_doc:
            for f in page.get_fonts(): xrefs.add(f[0])

        # process xref one by one
        fonts = []
        for xref in xrefs:
            basename, ext, _, buffer = fitz_doc.extract_font(xref)
            name = cls._normalized_font_name(basename)
            
            try:
                # supported fonts: open/true type only
                # - n/a: base 14 fonts
                # - cff: Adobe Compact File Format, i.e. Type 1 font
                assert ext not in ('n/a', 'cff'), "base font or not supported font"

                # try to get more font metrics with fonttool
                tt = TTFont(BytesIO(buffer))
                name = cls.get_font_family_name(tt)
                line_height = cls.get_line_height_factor(tt)
            except Exception:
                line_height = None

            fonts.append(Font(
                descriptor=cls._to_descriptor(name),
                name=name,
                line_height=line_height))

        return cls(fonts)
    

    @staticmethod
    def _normalized_font_name(name):
        '''Normalize raw font name, e.g. BCDGEE+Calibri-Bold, BCDGEE+Calibri -> Calibri.'''
        return name.split('+')[-1].split('-')[0]


    @staticmethod
    def _to_descriptor(name:str):
        '''Remove potential space, dash in font name, and turn to upper case.'''
        return name.replace(' ', '').replace('-', '').upper()

    
    @staticmethod
    def get_font_family_name(tt_font:TTFont):
        '''Get the font family name from the font's names table.

        https://gist.github.com/pklaus/dce37521579513c574d0
        '''
        name = family = ''
        FONT_SPECIFIER_NAME_ID = 4
        FONT_SPECIFIER_FAMILY_ID = 1

        for record in tt_font['name'].names:
            if b'\x00' in record.string:
                name_str = record.string.decode('utf-16-be')
            else:   
                name_str = record.string.decode('latin-1')

            if record.nameID == FONT_SPECIFIER_NAME_ID and not name:
                name = name_str
            elif record.nameID == FONT_SPECIFIER_FAMILY_ID and not family: 
                family = name_str

            if name and family: break

        # in case the font name is modified to pattern like BCDGEE+Calibri-Bold
        return Fonts._normalized_font_name(family)


    @staticmethod
    def get_line_height_factor(tt_font:TTFont):
        '''Calculate line height ratio based on ``hhea`` and ``OS/2`` tables.

        Fon non-CJK fonts::

            f = (hhea.Ascent - hhea.Descent + hhea.LineGap) / units_per_em
        
        For non-CJK fonts (Windows)::

            f = (OS/2.winAscent + OS/2.winDescent + [External Leading]) / units_per_em
            External Leading = MAX(0, hhea.LineGap - ((OS/2.WinAscent + OS/2.winDescent) - (hhea.Ascent - hhea.Descent)))

        For CJK fonts::

            f = 1.3 * (hhea.Ascent - hhea.Descent) / units_per_em

        Read more:
        * https://docs.microsoft.com/en-us/typography/opentype/spec/recom#baseline-to-baseline-distances
        * https://github.com/source-foundry/font-line#baseline-to-baseline-distance-calculations
        * https://www.zhihu.com/question/23349103
        * https://github.com/source-foundry/font-line/blob/master/lib/fontline/metrics.py
        '''
        units_per_em = tt_font["head"].unitsPerEm

        # hhea metrics
        hhea = tt_font["hhea"]
        hhea_ascent = hhea.ascent
        hhea_descent = hhea.descent
        hhea_linegap = hhea.lineGap

        hhea_total_height = hhea_ascent + abs(hhea_descent)
        hhea_btb_distance =  hhea_total_height + hhea_linegap

        # depends on System
        if os.name=='nt':

            # OS/2 metrics
            os2 = tt_font["OS/2"]
            os2_win_ascent = os2.usWinAscent
            os2_win_descent = os2.usWinDescent
            os2_win_total_height = os2_win_ascent + os2_win_descent
            win_external_leading = max(0.0, hhea_linegap-(os2_win_total_height-hhea_total_height))
            win_btb_distance = os2_win_total_height + win_external_leading
            
            btb_distance = win_btb_distance
        
        else:
            btb_distance = hhea_btb_distance

        # depends on CJK font or not
        cjk = Fonts.is_cjk_font(tt_font)
        distance = 1.3*hhea_total_height if cjk else 1.0*btb_distance

        return distance / units_per_em
    

    @staticmethod
    def is_cjk_font(tt_font:TTFont):
        '''Test font object to confirm that it meets our definition of a CJK font file.

        The definition is met if any of the following conditions are True:
        1. The font has a CJK code page bit set in the OS/2 table
        2. The font has a CJK Unicode range bit set in the OS/2 table
        3. The font has any CJK Unicode code points defined in the cmap table

        https://github.com/googlefonts/fontbakery/blob/main/Lib/fontbakery/profiles/shared_conditions.py
        '''
        os2 = tt_font["OS/2"]

        # OS/2 code page checks
        for _, bit in CJK_CODEPAGE_BITS.items():
            if hasattr(os2, 'ulCodePageRange1') and os2.ulCodePageRange1 & (1 << bit):
                return True

        # OS/2 Unicode range checks
        for _, bit in CJK_UNICODE_RANGE_BITS.items():
            if bit in range(0, 32):
                if hasattr(os2, 'ulCodePageRange1') and os2.ulUnicodeRange1 & (1 << bit):
                    return True

            elif bit in range(32, 64):
                if hasattr(os2, 'ulCodePageRange2') and os2.ulUnicodeRange2 & (1 << (bit-32)):
                    return True

            elif bit in range(64, 96):
                if hasattr(os2, 'ulCodePageRange3') and os2.ulUnicodeRange3 & (1 << (bit-64)):
                    return True

        # defined CJK Unicode code point in cmap table checks
        try:
            cmap = tt_font.getBestCmap()
        except:
            return False
        if not cmap: return False

        for unicode_range in CJK_UNICODE_RANGES:
            for x in range(unicode_range[0], unicode_range[1]+1):
                if int(x) in cmap:
                    return True

        # default, return False if the above checks did not identify a CJK font
        return False



# -*- coding: utf-8 -*-

'''
Objects representing PDF path (stroke and filling) extracted from pdf drawings and annotations.

Data structure based on results of ``page.get_drawings()``::

    {
        'color': (x,x,x) or None,  # stroke color
        'fill' : (x,x,x) or None,  # fill color
        'width': float,            # line width
        'closePath': bool,         # whether to connect last and first point
        'rect' : rect,             # page area covered by this path
        'items': [                 # list of draw commands: lines, rectangle or curves.
            ("l", p1, p2),         # a line from p1 to p2
            ("c", p1, p2, p3, p4), # cubic Bézier curve from p1 to p4, p2 and p3 are the control points
            ("re", rect),          # a rect represented with two diagonal points
            ("qu", quad)           # a quad represented with four corner points
        ],
        ...
    }

References: 
    - https://pymupdf.readthedocs.io/en/latest/page.html#Page.get_drawings
    - https://pymupdf.readthedocs.io/en/latest/faq.html#extracting-drawings

.. note::
    The coordinates extracted by ``page.get_drawings()`` is based on **real** page CS, i.e. with rotation 
    considered. This is different from ``page.get_text('rawdict')``.
'''

import fitz
from ..common.share import rgb_value
from ..common import constants


class Segment:
    '''A segment of path, e.g. a line or a rectangle or a curve.'''
    def __init__(self, item):
        self.points = item[1:]

    def to_strokes(self, width:float, color:list): return []


class L(Segment):
    '''Line path with source ``("l", p1, p2)``.'''

    @property
    def length(self):
        x0, y0 = self.points[0]
        x1, y1 = self.points[1]
        return ((x1-x0)**2+(y1-y0)**2)**0.5


    def to_strokes(self, width:float, color:list):
        """Convert to stroke dict.

        Args:
            width (float): Specify width for the stroke.
            color (list): Specify color for the stroke.

        Returns:
            list: A list of ``Stroke`` dicts. 
        
        .. note::
            A line corresponds to one stroke, but considering the consistence, 
            the return stroke dict is append to a list. So, the length of list 
            is always 1.
        """ 
        strokes = []
        strokes.append({
                'start': self.points[0],
                'end'  : self.points[1],
                'width': width,
                'color': rgb_value(color)
            })
        return strokes


class R(Segment):
    '''Rect path with source ``("re", rect)``.'''
    def __init__(self, item):
        # corner points
        # NOTE: center line of path without stroke width considered
        x0, y0, x1, y1 = item[1]
        self.points = [
            (x0, y0), (x1, y0), (x1, y1), (x0, y1), (x0, y0) # close path
            ]


    def to_strokes(self, width:float, color:list):
        """Convert each edge to stroke dict.

        Args:
            width (float): Specify width for the stroke.
            color (list): Specify color for the stroke.

        Returns:
            list: A list of ``Stroke`` dicts. 
        
        .. note::
            One Rect path is converted to a list of 4 stroke dicts.
        """
        strokes = []
        for i in range(len(self.points)-1):
            strokes.append({
                    'start': self.points[i],
                    'end'  : self.points[i+1],
                    'width': width * 2.0, # seems need adjustment by * 2.0
                    'color': rgb_value(color)
                })
        return strokes


class Q(R):
    '''Quad path with source ``("qu", quad)``.'''
    def __init__(self, item):
        # four corner points
        # NOTE: center line of path without stroke width considered
        ul, ur, ll, lr = item[1]
        self.points = [ul, ur, lr, ll, ul] # close path


class C(Segment):
    '''Bezier curve path with source ``("c", p1, p2, p3, p4)``.'''
    pass


class Segments:
    '''A sub-path composed of one or more segments.'''
    def __init__(self, items:list, close_path=False): 
        self._instances = [] # type: list[Segment]
        for item in items:
            if   item[0] == 'l' : self._instances.append(L(item))
            elif item[0] == 'c' : self._instances.append(C(item))
            elif item[0] == 're': self._instances.append(R(item))
            elif item[0] == 'qu': self._instances.append(Q(item))
        
        # close path
        if close_path:
            item = ('l', self._instances[-1].points[-1], self._instances[0].points[0])
            line = L(item)
            if line.length>1e-3: self._instances.append(line)

    
    def __iter__(self): return (instance for instance in self._instances)


    @property
    def points(self):
        '''Connected points of segments.'''
        points = []
        for segment in self._instances:
            points.extend(segment.points)
        return points


    @property
    def is_iso_oriented(self):
        '''ISO-oriented criterion: the ratio of real area to bbox exceeds 0.9.'''
        bbox_area = self.bbox.get_area()
        return bbox_area==0 or self.area/bbox_area>=constants.FACTOR_MOST


    @property
    def area(self):
        '''Calculate segments area with Green formulas. Note the boundary of Bezier curve 
        is simplified with its control points.
        
        * https://en.wikipedia.org/wiki/Shoelace_formula
        '''        
        points = self.points
        start, end = points[0], points[-1]
        if abs(start[0]-end[0])+abs(start[1]-end[1])>1e-3: 
            return 0.0 # open curve
            
        # closed curve 
        area = 0.0       
        for i in range(len(points)-1):
            x0, y0 = points[i]
            x1, y1 = points[i+1]
            area += x0*y1 - x1*y0

        return abs(area/2.0)


    @property
    def bbox(self):
        '''Calculate segments bbox. '''
        points = self.points
        x0 = min(points, key=lambda point: point[0])[0]
        y0 = min(points, key=lambda point: point[1])[1]
        x1 = max(points, key=lambda point: point[0])[0]
        y1 = max(points, key=lambda point: point[1])[1]
        
        # bbox: `round()` is required to avoid float error
        return fitz.Rect(
            round(x0, 2), round(y0, 2), round(x1, 2), round(y1, 2))


    def to_strokes(self, width:float, color:list):
        """Convert each segment to a ``Stroke`` dict.

        Args:
            width (float): Specify stroke width.
            color (list): Specify stroke color.

        Returns:
            list: A list of ``Stroke`` dicts.
        """        
        strokes = []
        for segment in self._instances: 
            strokes.extend(segment.to_strokes(width, color))
        return strokes
    

    def to_fill(self, color:list):
        """Convert segment closed area to a ``Fill`` dict.

        Args:
            color (list): Specify fill color.

        Returns:
            dict: ``Fill`` dict.
        """        
        return {
            'bbox' : list(self.bbox), 
            'color': rgb_value(color)
        }


class Path:
    '''Path extracted from PDF, consist of one or more ``Segments``.'''
    def __init__(self, raw:dict):
        '''Init path in real page CS.

        Args:
            raw (dict): Raw dict extracted with `PyMuPDF`, see link
            https://pymupdf.readthedocs.io/en/latest/page.html#Page.get_drawings
        '''
        # all path properties
        self.raw = raw
        self.path_type = raw['type'] # s, f, or fs

        # always close path if fill, otherwise, depends on property 'closePath'
        close_path = True if self.is_fill else raw.get('closePath', False)

        # path segments
        self.items = [] # type: list[Segments]
        self.bbox = fitz.Rect()
        w = raw.get('width', 0.0)
        for segments in self._group_segments(raw['items']):
            S = Segments(segments, close_path)
            self.items.append(S)

            # update bbox: note iso-oriented line segments -> S.bbox.get_area()==0
            rect = S.bbox
            if rect.get_area()==0: rect += (-w, -w, w, w)
            self.bbox |= rect


    @staticmethod
    def _group_segments(items):
        """Group connected segments.

        Args:
            items (dict): Raw dict extracted from ``page.get_drawings()``.

        Returns:
            list: A list of segments list.
        """        
        segments, segments_list = [], []
        cursor = None
        for item in items:
            # line or curve segment
            if item[0] in ('l', 'c'):
                start, end = item[1], item[-1]
                # add to segments if:
                # - first point of segments, or
                # - connected to previous segment
                if not segments or start==cursor:
                    segments.append(item)                    
                
                # otherwise, close current segments and start a new one
                else:
                    segments_list.append(segments)
                    segments = [item] 
                
                # update current point
                cursor = end

            # rectangle / quad as a separate segments group
            elif item[0] in ('re', 'qu'):
                # close current segments
                if segments:
                    segments_list.append(segments)
                    segments = []
                # add this segment
                segments_list.append([item])
        
        # add last segments if exists
        if segments: segments_list.append(segments)

        return segments_list
                

    @property
    def is_stroke(self): return 's' in self.path_type

    @property
    def is_fill(self): return 'f' in self.path_type

    @property
    def is_iso_oriented(self):
        '''It is iso-oriented when all contained segments are iso-oriented.'''
        for segments in self.items:
            if not segments.is_iso_oriented: return False
        return True


    def to_shapes(self):
        """Convert path to ``Shape`` raw dicts.

        Returns:
            list: A list of ``Shape`` dict.
        """
        iso_shapes = []

        # convert to strokes
        if self.is_stroke:
            stroke_color = self.raw.get('color', None)
            width = self.raw.get('width', 0.0)
            iso_shapes.extend(self._to_strokes(width, stroke_color))

        # convert to rectangular fill
        if self.is_fill:
            fill_color = self.raw.get('fill', None)
            iso_shapes.extend(self._to_fills(fill_color))

        return iso_shapes


    def _to_strokes(self, width:float, color:list):
        '''Convert path to ``Stroke`` raw dicts.

        Returns:
            list: A list of ``Stroke`` dict.
        '''
        strokes = []        
        for segments in self.items:
            strokes.extend(segments.to_strokes(width, color))        
        return strokes


    def _to_fills(self, color:list):
        '''Convert path to ``Fill`` raw dicts.

        Returns:
            list: A list of ``Fill`` dict.
        
        .. note::
            The real filling area of this path may be not a rectangle.        
        '''
        fills = []        
        for segments in self.items:
            fills.append(segments.to_fill(color))        
        return fills


    def plot(self, canvas):
        ''' Plot path for debug purpose.

        Args:
            canvas: ``PyMuPDF`` drawing canvas by ``page.new_shape()``.

        Reference:
        
            https://pymupdf.readthedocs.io/en/latest/faq.html#extracting-drawings
        '''
        # draw each entry of the 'items' list
        for item in self.raw.get('items', []):
            if item[0] == "l":  # line
                canvas.draw_line(item[1], item[2])
            elif item[0] == "re":  # rectangle
                canvas.draw_rect(item[1])
            elif item[0] == "qu":  # quad
                canvas.draw_quad(item[1])
            elif item[0] == "c":  # curve
                canvas.draw_bezier(item[1], item[2], item[3], item[4])
            else:
                raise ValueError("unhandled drawing", item)

        # now apply the common properties to finish the path
        canvas.finish(
            fill=self.raw.get("fill", None),  # fill color
            color=self.raw.get("color", None),  # line color
            dashes=self.raw.get("dashes", None),  # line dashing
            even_odd=self.raw.get("even_odd", False),  # control color of overlaps
            closePath=self.raw.get("closePath", False),  # whether to connect last and first point
            lineJoin=self.raw.get("lineJoin", 0),  # how line joins should look like
            lineCap=max(self.raw["lineCap"]) if "lineCap" in self.raw else 0,  # how line ends should look like
            width=self.raw.get("width", 1),  # line width
            stroke_opacity=self.raw.get("stroke_opacity", 1),  # same value for both
            fill_opacity=self.raw.get("fill_opacity", 1)  # opacity parameters
            )


#!/usr/bin/python3
# -*- coding: utf-8 -*-

import logging
from .converter import Converter


class pdf2word:
    '''Command line interface for ``pdf2word``.'''

    @staticmethod
    def convert(pdf_file:str, docx_file:str=None, password:str=None, start:int=0, end:int=None, pages:list=None, **kwargs):
        '''Convert pdf file to docx file.
        
        Args:
            pdf_file (str) : PDF filename to read from.
            docx_file (str, optional): docx filename to write to. Defaults to None.
            password (str): Password for encrypted pdf. Default to None if not encrypted.
            start (int, optional): First page to process. Defaults to 0.
            end (int, optional): Last page to process. Defaults to None.
            pages (list, optional): Range of pages, e.g. --pages=1,3,5. Defaults to None.
            kwargs (dict) : Configuration parameters.
        
        .. note::
            Refer to :py:meth:`~pdf2word.converter.Converter.convert` for detailed description on above arguments.
        '''
        # index starts from zero or one
        if isinstance(pages, int): pages = [pages] # in case --pages=1
        if not kwargs.get('zero_based_index', True):
            start = max(start-1, 0)
            if end: end -= 1
            if pages: pages = [i-1 for i in pages]

        cv = Converter(pdf_file, password)
        try:
            cv.convert(docx_file, start, end, pages, **kwargs)
        except Exception as e:
            logging.error(e)
        finally:
            cv.close()
    

    @staticmethod
    def debug(pdf_file:str, password:str=None, page:int=0, docx_file:str=None, debug_pdf:str=None, layout_file:str='layout.json', **kwargs):
        '''Convert one PDF page and plot layout information for debugging.
        
        Args:
            pdf_file (str) : PDF filename to read from.
            password (str): Password for encrypted pdf. Default to None if not encrypted.
            page (int, optional): Page index to convert.
            docx_file (str, optional): docx filename to write to.
            debug_pdf (str, optional): Filename for new pdf storing layout information. Defaults to same name with pdf file.
            layout_file (str, optional): Filename for new json file storing parsed layout data. Defaults to ``layout.json``.
            kwargs (dict)  : Configuration parameters.
        '''
        # index starts from zero or one
        if not kwargs.get('zero_based_index', True):
            page = max(page-1, 0)

        # explode exception directly if debug mode
        cv = Converter(pdf_file, password)
        cv.debug_page(page, docx_file, debug_pdf, layout_file, **kwargs)
        cv.close()
            


    @staticmethod
    def table(pdf_file, password:str=None, start:int=0, end:int=None, pages:list=None, **kwargs):
        '''Extract table content from pdf pages.
        
        Args:
            pdf_file (str) : PDF filename to read from.
            password (str): Password for encrypted pdf. Default to None if not encrypted.
            start (int, optional): First page to process. Defaults to 0.
            end (int, optional): Last page to process. Defaults to None.
            pages (list, optional): Range of pages, e.g. --pages=1,3,5. Defaults to None.
        '''
        # index starts from zero or one
        if isinstance(pages, int): pages = [pages] # in case --pages=1
        if not kwargs.get('zero_based_index', True):
            start = max(start-1, 0)
            if end: end -= 1
            if pages: pages = [i-1 for i in pages]
        
        cv = Converter(pdf_file, password)
        try:
            tables = cv.extract_tables(start, end, pages, **kwargs)
        except Exception as e:
            tables = []
            logging.error(e)
        finally:
            cv.close()

        return tables


    @staticmethod
    def gui():
        '''Simple user interface.'''
        # import App containing tkinter internally, in case GUI is not supported by some platforms,
        # e.g. Amazon Linux 2
        from .gui.App import App
        app = App(title='PDF_2_Docx Converter', width=500, height=600)
        app.mainloop()


parse = pdf2word.convert


def main():
    import fire
    fire.Fire(pdf2word)


if __name__ == '__main__':
    main()


# -*- coding: utf-8 -*-

'''Definition of Image block objects. 

**The raw image block will be merged into TextBlock > Line > Span.**
'''

from io import BytesIO
from ..text.Line import Line
from ..text.TextBlock import TextBlock
from .Image import Image
from .ImageSpan import ImageSpan
from ..common.Block import Block
from ..common.docx import add_float_image


class ImageBlock(Image, Block):
    '''Image block.'''
    def __init__(self, raw:dict=None):
        super().__init__(raw)

        # inline image type by default
        self.set_inline_image_block()


    def to_text_block(self):
        """Convert image block to a span under text block.

        Returns:
            TextBlock: New TextBlock instance containing this image.
        """
        # image span
        span = ImageSpan().from_image(self)

        # add span to line
        image_line = Line()
        image_line.add(span)
        
        # insert line to block
        block = TextBlock()        
        block.add(image_line)

        # NOTE: it's an image block even though in TextBlock type
        block.set_inline_image_block() 

        return block
 

    def store(self):
        '''Store ImageBlock instance in raw dict.'''
        res = Block.store(self)
        res.update(
            Image.store(self)
        )
        return res

    
    def plot(self, page):
        '''Plot image bbox with diagonal lines (for debug purpose).
        
        Args: 
            page (fitz.Page): pdf page to plot.
        '''
        super().plot(page, color=(1,0,0))


    def make_docx(self, p):
        '''Create floating image behind text. 
        
        Args:
            p (Paragraph): ``python-docx`` paragraph instance.
        
        .. note::
            Inline image is created within TextBlock.
        '''
        if self.is_float_image_block:
            x0, y0, x1, y1 = self.bbox
            add_float_image(p, BytesIO(self.image), width=x1-x0, pos_x=x0, pos_y=y0)
        else:
            super().make_docx(p)
        return p


# -*- coding: utf-8 -*-

'''Objects representing PDF stroke and filling extracted from Path.

* Stroke: consider only the horizontal or vertical path segments
* Fill  : bbox of closed path filling area

Hyperlink in ``PyMuPDF`` is represented as uri and its rectangular area (hot-area), while the
applied text isn't extracted explicitly. To reuse the process that identifying applied text of
text style shape (e.g. underline and highlight), hyperlink is also abstracted to be a ``Shape``.

.. note::
    The evident difference of hyperlink shape to text style shape is: the ``type`` of hyperlink 
    shape is determined in advance, while text style shape needs to be identified by the position 
    to associated text blocks.

Above all, the semantic meaning of shape instance may be:

* strike through line of text
* under line of text
* highlight area of text
* table border
* cell shading
* hyperlink

Data structure::

    {
        'type': int,
        'bbox': (x0, y0, x1, y1),
        'color': srgb_value,

        # for Stroke
        'start': (x0, y0),
        'end': (x1, y1),
        'width': float,

        # for Hyperlink
        'uri': str
    }

.. note::
    These coordinates are relative to real page CS since they're extracted from ``page.get_drawings()``,
    which is based on real page CS. So, needn't to multiply Element.ROTATION_MATRIX when initializing
    from source dict.
'''

import fitz
from ..common.Element import Element
from ..common.share import RectType
from ..common import constants


class Shape(Element):
    ''' Shape object.'''
    def __init__(self, raw:dict=None):        
        raw = raw or {}
        self.color = raw.get('color', 0)
        
        # NOTE: coordinates are based on real page CS already
        super().update_bbox(raw.get('bbox', (0,)*4))
        self._parent = None

        # shape semantic type
        # It's able to set shape type in ``RectType``, but a shape might belong to multi-types before
        # it's finalized. So, set int type here.
        self._type = raw.get('type', -1)            # final type
        self._potential_type = self.default_type    # potential types, a combination of raw RectType-s


    @property
    def type(self): return self._type

    @type.setter
    def type(self, rect_type:RectType): self._type = rect_type.value

    @property
    def is_determined(self): 
        '''If the shape type is determined to a basic item of RectType.'''
        return self._type != -1

    def equal_to_type(self, rect_type:RectType): 
        '''If shape type is equal to the specified one or not.'''
        return self._type==rect_type.value

    def has_potential_type(self, rect_type:RectType):
        '''If shape type has a chance to be the specified one or not.'''
        return self._potential_type & rect_type.value

    @property
    def default_type(self): 
        '''Default semantic type for a shape.'''
        return sum(t.value for t in RectType)


    def store(self):
        res = super().store()
        res.update({
            'type': self._type,
            'color': self.color
        })
        return res


    def parse_semantic_type(self, blocks:list):
        '''Determine semantic type based on the position to text blocks. Note the results might be 
        a combination of raw types, e.g. the semantic type of a stroke can be either text strike,
        underline or table border.

        Args:
            blocks (list): A list of ``Line`` instance, sorted in reading order in advance.
        '''
        for line in blocks:
            # not intersect yet
            if line.bbox.y1 < self.bbox.y0: continue
            
            # no intersection any more
            if line.bbox.y0 > self.bbox.y1: break

            # check it when intersected
            rect_type = self._semantic_type(line)
            self._potential_type = rect_type

            if rect_type!=self.default_type: break


    def _semantic_type(self, line):
        ''' Check semantic type based on the position to a text line.
            Return all possibilities if can't be determined with this text line.
            Prerequisite: intersection exists between this shape and line.
        '''
        return self.default_type


    def plot(self, page, color):
        '''Plot rectangle shapes with ``PyMuPDF``.'''
        page.draw_rect(self.bbox, color=color, fill=color, width=0, overlay=True)


class Stroke(Shape):
    ''' Horizontal or vertical stroke of a path. 
        The semantic meaning may be table border, or text style line like underline and strike-through.
    '''
    def __init__(self, raw:dict=None):
        raw = raw or {}
        # NOTE: real page CS
        self._start = fitz.Point(raw.get('start', (0.0, 0.0)))
        self._end = fitz.Point(raw.get('end', (0.0, 0.0)))

        if self._start.x > self._end.x or self._start.y > self._end.y:
            self._start, self._end = self._end, self._start

        # width, color
        super().__init__(raw) # type, color
        self.width = raw.get('width', 0.0) # Note this "width" is actually the height of stroke

        # update bbox
        super().update_bbox(self._to_rect())


    @property
    def horizontal(self): return abs(self._start[1]-self._end[1])<1e-3

    @property
    def vertical(self): return abs(self._start[0]-self._end[0])<1e-3

    @property
    def x0(self): return self._start.x

    @property
    def x1(self): return self._end.x

    @property
    def y0(self): return self._start.y

    @property
    def y1(self): return self._end.y


    def update_bbox(self, rect):
        '''Update stroke bbox (related to real page CS).

        * Update start/end points if ``rect.area==0``.
        * Ppdate bbox directly if ``rect.area!=0``.

        Args:
            rect (fitz.Rect, tuple): ``(x0, y0, x1, y1)`` like data.

        Returns:
            Stroke: self
        '''
        rect = fitz.Rect(rect)

        # an empty area line
        if rect.get_area()==0.0:
            self._start = fitz.Point(rect[0:2])
            self._end = fitz.Point(rect[2:])
            super().update_bbox(self._to_rect())

        # a rect 
        else:
            super().update_bbox(rect)

            # horizontal stroke
            if rect.width >= rect.height:
                y = (rect.y0+rect.y1)/2.0
                self._start = fitz.Point(rect.x0, y)
                self._end   = fitz.Point(rect.x1, y)

            # vertical stroke
            else: 
                x = (rect.x0+rect.x1)/2.0
                self._start = fitz.Point(x, rect.y0)
                self._end   = fitz.Point(x, rect.y1)

        return self

    @property
    def default_type(self):
        '''Default semantic type for a Stroke shape: table border, underline or strike-through.'''
        return RectType.BORDER.value | RectType.UNDERLINE.value | RectType.STRIKE.value

    def _semantic_type(self, line):
        '''Override. Check semantic type of a Stroke: table border v.s. text style line, e.g. underline 
        and strike-through. It's potentially a text style line when:

        * the stroke and the text line has same orientation; and
        * the stroke never exceeds the text line along the main direction
        '''
        # check intersection
        expanded_shape = self.get_expand_bbox(2.0)
        if not line.bbox.intersects(expanded_shape): 
            return self.default_type        

        # check orientation
        h_shape = self.horizontal
        h_line = line.is_horizontal_text
        if h_shape != h_line: 
            return self.default_type

        # check main dimension
        line_x0, line_x1 = (line.bbox.x0, line.bbox.x1) if h_line else (line.bbox.y0, line.bbox.y1)
        shape_x0, shape_x1 = (self.bbox.x0, self.bbox.x1) if h_shape else (self.bbox.y0, self.bbox.y1)
        if shape_x0>=line_x0-1 and shape_x1<=line_x1+1: # 1 pt tolerance at both sides
            return RectType.STRIKE.value | RectType.UNDERLINE.value
        else:
            return RectType.BORDER.value        


    def store(self):
        res = super().store()
        res.update({
            'start': tuple(self._start),
            'end': tuple(self._end),
            'width': self.width
        })
        return res


    def _to_rect(self):
        '''Convert centerline to rectangle shape.'''
        h = self.width / 2.0
        x0, y0 = self._start
        x1, y1 = self._end        
        return (x0-h, y0-h, x1+h, y1+h)


class Fill(Shape):
    ''' Rectangular (bbox) filling area of a closed path. 
        The semantic meaning may be table shading, or text style like highlight.
    '''

    def to_stroke(self, max_border_width:float):
        '''Convert to Stroke instance based on width criterion.

        Args:
            max_border_width (float): Stroke width must less than this value.

        Returns:
            Stroke: Stroke instance.
        
        .. note::
            A Fill from shape point of view may be a Stroke from content point of view.
            The criterion here is whether the width is smaller than defined ``max_border_width``.
        '''
        w = min(self.bbox.width, self.bbox.height)

        # not a stroke if exceed max border width
        if w > max_border_width:
            return None
        else:
            return Stroke({'width': w, 'color': self.color}).update_bbox(self.bbox)
    

    @property
    def default_type(self):
        '''Default semantic type for a Fill shape: table shading or text highlight.'''
        return RectType.SHADING.value | RectType.HIGHLIGHT.value
    
    def _semantic_type(self, line):
        '''Override. Check semantic type based on the position to a text line. Along the main dimension,
        text highlight never exceeds text line.

        Args:
            line (Line): A text line.

        Returns:
            RectType: Semantic type of this shape.
        
        .. note::
            Generally, table shading always contains at least one line, while text highlight never
            contains any lines. But in real cases, with margin exists, table shading may not 100% 
            contain a line.
        '''
        # check main dimension
        h_shape = self.bbox.width>self.bbox.height
        w_shape = self.bbox.width if h_shape else self.bbox.height

        # check orientation
        h_line = line.is_horizontal_text
        if h_shape != h_line: 
            return self.default_type

        if not self.get_main_bbox(line, threshold=constants.FACTOR_MAJOR): 
            return self.default_type
        
        w_line = line.bbox.width if h_line else line.bbox.height            
        if w_shape <= w_line + 2*constants.MINOR_DIST: # 1 pt tolerance at both sides
            return RectType.HIGHLIGHT.value
        else:
            return RectType.SHADING.value        


class Hyperlink(Shape):
    '''Rectangular area, i.e. ``hot area`` for a hyperlink. 
    
    Hyperlink in ``PyMuPDF`` is represented as uri and its hot area, while the applied text isn't extracted 
    explicitly. To reuse the process that identifying applied text of text style shape (e.g. underline and 
    highlight), hyperlink is also abstracted to be a ``Shape``.
    '''

    def __init__(self, raw:dict=None):
        '''Initialize from raw dict. Note the type must be determined in advance.'''
        super().__init__(raw)

        # set uri
        self.uri = raw.get('uri', '')
        

    def store(self):
        res = super().store()
        res.update({
            'uri': self.uri
        })
        return res


    @property
    def default_type(self):
        '''Default semantic type for a Hyperlink: always hyperlink.'''
        return RectType.HYPERLINK.value

    def parse_semantic_type(self, blocks:list=None):
        '''Semantic type of Hyperlink shape is determined, i.e. ``RectType.HYPERLINK``.'''
        self._potential_type = self.default_type


# -*- coding: utf-8 -*-

'''A group of TextSpan and ImageSpan objects.
'''

from ..common.Collection import ElementCollection
from .TextSpan import TextSpan
from ..image.ImageSpan import ImageSpan

class Spans(ElementCollection):
    '''Collection of TextSpan and ImageSpan instances.'''

    def restore(self, raws:list):
        '''Recreate TextSpan or ImageSpan from source dict list.'''
        for raw_span in raws:
            if 'image' in raw_span:
                span = ImageSpan(raw_span)
            else:
                span = TextSpan(raw_span)
                if not span.text.strip() and not span.style: 
                    span = None

            self.append(span)
        return self

    @property
    def text_spans(self):
        '''Get TextSpan instances.'''
        spans = list(filter(
            lambda span: isinstance(span, TextSpan), self._instances
        ))
        return Spans(spans)

    @property
    def image_spans(self):
        '''Get ImageSpan instances.'''
        spans = list(filter(
            lambda span: isinstance(span, ImageSpan), self._instances
        ))
        return Spans(spans)


    def strip(self):
        '''Remove redundant blanks at the begin/end span.'''
        stripped = False
        if not self._instances: return stripped
        
        # left strip the first span
        left_span = self._instances[0]
        if isinstance(left_span, TextSpan): stripped = stripped or left_span.lstrip() 

        # right strip the last span
        right_span = self._instances[-1]
        if isinstance(right_span, TextSpan): stripped = stripped or right_span.rstrip()

        # update bbox
        if stripped: self._parent.update_bbox(self.bbox)

        return stripped


# -*- coding: utf-8 -*-

'''
Objects representing PDF path (stroke and filling) extracted by ``page.get_drawings()``.

This method is new since ``PyMuPDF`` 1.18.0, with both pdf raw path and annotations like Line, 
Square and Highlight considered.

* https://pymupdf.readthedocs.io/en/latest/page.html#Page.get_drawings
* https://pymupdf.readthedocs.io/en/latest/faq.html#extracting-drawings
'''

import fitz
from ..image.ImagesExtractor import ImagesExtractor
from ..common.share import lazyproperty
from ..common.Collection import  Collection
from .Path import Path


class Paths(Collection):
    '''A collection of paths.'''

    def restore(self, raws:list):
        '''Initialize paths from raw data get by ``page.get_drawings()``.'''
        rect = (0, 0, self.parent.width, self.parent.height)
        for raw in raws:
            path = Path(raw)
            # ignore path out of page
            if not path.bbox.intersects(rect): continue
            self.append(path)
        
        return self
    
    @lazyproperty
    def bbox(self):
        bbox = fitz.Rect()
        for instance in self._instances: bbox |= instance.bbox
        return bbox


    @property
    def is_iso_oriented(self):
        '''It is iso-oriented when all contained segments are iso-oriented.'''
        for instance in self._instances:
            if not instance.is_iso_oriented: return False
        return True


    def plot(self, page):
        '''Plot paths for debug purpose.

        Args:
            page (fitz.Page): ``PyMuPDF`` page.
        '''
        if not self._instances: return
        # make a drawing canvas and plot path
        canvas = page.new_shape()
        for path in self._instances: path.plot(canvas)
        canvas.commit() # commit the drawing shapes to page
    

    def to_shapes(self):
        '''Convert contained paths to ISO strokes or rectangular fills.

        Returns:
            list: A list of ``Shape`` raw dicts.
        '''
        shapes = []
        for path in self._instances:
            # consider iso-oriented path only
            if not path.is_iso_oriented: continue
            shapes.extend(path.to_shapes())
        return shapes


    def to_shapes_and_images(self, min_svg_gap_dx:float=15, min_svg_gap_dy:float=15, 
                                min_w:float=2, min_h:float=2, clip_image_res_ratio:float=3.0):
        '''Convert paths to iso-oriented shapes or images. The semantic type of path is either table/text style or 
        vector graphic. This method is to:
        * detect svg regions -> exist at least one non-iso-oriented path
        * convert svg to bitmap by clipping page
        * convert the rest paths to iso-oriented shapes for further table/text style parsing

        Args:
            min_svg_gap_dx (float): Merge svg if the horizontal gap is less than this value.
            min_svg_gap_dy (float): Merge svg if the vertical gap is less than this value.
            min_w (float): Ignore contours if the bbox width is less than this value.
            min_h (float): Ignore contours if the bbox height is less than this value.
            clip_image_res_ratio (float, optional): Resolution ratio of clipped bitmap. Defaults to 3.0.

        Returns:
            tuple: (list of shape raw dict, list of image raw dict).
        '''
        # convert all paths to shapes if no non-iso-orientated path exists
        iso_shapes = []
        if self.is_iso_oriented:
            iso_shapes.extend(self.to_shapes())
            return iso_shapes, []

        # detect svg with python opencv
        images = []
        ie = ImagesExtractor(self.parent.page_engine)
        groups = ie.detect_svg_contours(min_svg_gap_dx, min_svg_gap_dy, min_w, min_h)

        # `bbox` is the external bbox of current region, while `inner_bboxes` are the inner contours
        # of level-2 hierarchy, i.e. contours under table cell.
        # * it a table (or text style) if paths contained in `bbox` but excluded from `inner_bboxes` 
        #   are all iso-oriented -> export iso-shapes, clip page image based on `inner_bboxes`;
        # * otherwise, it's a vector graphic -> clip page image (without any text) based on `bbox`
        def contained_in_inner_contours(path:Path, contours:list):
            for bbox in contours:
                if fitz.Rect(bbox).contains(path.bbox): return True
            return False

        # group every path to one of the detected bbox
        group_paths = [Paths() for _ in groups] # type: list[Paths]
        for path in self._instances:
            for (bbox, inner_bboxes), paths in zip(groups, group_paths):            
                if path.bbox.intersects(bbox):
                    if not contained_in_inner_contours(path, inner_bboxes): paths.append(path)
                    break
        
        # check each group
        for (bbox, inner_bboxes), paths in zip(groups, group_paths): 
            # all iso-oriented paths -> it's a table, but might contain svg in cell as well
            if paths.is_iso_oriented:
                iso_shapes.extend(paths.to_shapes())
                for svg_bbox in inner_bboxes:
                    images.append(ie.clip_page_to_dict(fitz.Rect(svg_bbox), clip_image_res_ratio))
            
            # otherwise, it's a svg
            else:
                images.append(ie.clip_page_to_dict(fitz.Rect(bbox), clip_image_res_ratio))

        return iso_shapes, images




# -*- coding: utf-8 -*-

'''Parsing table structure based on strokes and fills.
'''

import fitz
from ..common.Element import Element
from ..common.share import RectType
from ..common import constants
from ..shape.Shape import Shape, Stroke
from ..shape.Shapes import Shapes
from .TableBlock import TableBlock
from .Row import Row
from .Cell import Cell


class CellStructure:
    '''Cell structure with properties bbox, borders, shading, etc.'''
    def __init__(self, bbox:list):
        # bbox
        self.bbox = fitz.Rect(bbox) # theoretical lattice bbox
        self.merged_bbox = fitz.Rect(bbox) # cell bbox considering merged cells

        # stroke shapes around this cell: top, right, bottom, left
        self.borders = None # type: list[Shape]
        
        # fill shape representing the bg-color
        self.shading = None # type: Shape

        # the count of merged cells in row and col direction:
        # (1, 1) by default -> the cell itself -> no merged cell
        # (3, 2) -> merge 3*2=6 cells (with itself counted as the top-left cell)
        # (0, 0) -> it is merged by other cell
        self.merged_cells = (1,1)

   
    @property
    def is_merged(self): return self.merged_cells[0]==0 or self.merged_cells[1]==0

    @property
    def is_merging(self): return self.merged_cells[0]>1 or self.merged_cells[1]>1


    def parse_borders(self, h_strokes:dict, v_strokes:dict):
        '''Parse cell borders from strokes.
        
        Args:
            h_strokes (dict): A dict of y-coordinate v.s. horizontal strokes, e.g. 
                ``{y0: [h1,h2,..], y1: [h3,h4,...]}``
            v_strokes (dict): A dict of x-coordinates v.s. vertical strokes, e.g. 
                ``{x0: [v1,v2,..], x1: [v3,v4,...]}``
        '''
        x0, y0, x1, y1 = self.merged_bbox
        top = self._get_border_stroke(h_strokes[y0], 'row')
        bottom = self._get_border_stroke(h_strokes[y1], 'row')
        left = self._get_border_stroke(v_strokes[x0], 'col')
        right = self._get_border_stroke(v_strokes[x1], 'col')
        self.borders = (top, bottom, left, right)

    
    def parse_shading(self, fills:Shapes):
        '''Parse cell shading from fills.
        
        Args:
            fills (Shapes): Fill shapes representing cell shading.
        '''
        # border width
        top, bottom, left, right = self.borders
        w_top = top.width
        w_right = right.width
        w_bottom = bottom.width
        w_left = left.width

        # modify the cell bbox from border center to inner region
        x0, y0, x1, y1 = self.merged_bbox
        inner_bbox = (x0+w_left/2.0, y0+w_top/2.0, x1-w_right/2.0, y1-w_bottom/2.0)
        target = Element().update_bbox(inner_bbox)

        # shading shape of this cell        
        for shape in fills:
            if shape.contains(target, threshold=constants.FACTOR_MOST):
                self.shading = shape
                break
        else:
            self.shading = None


    def _get_border_stroke(self, strokes:Shapes, direction:str='row'):
        ''' Find strokes representing cell borders.
        
        Args:
            strokes (Shapes): Candidate stroke shapes for cell border.
            direction (str): Either ``row`` or ``col``.
        '''
        if not strokes: return Stroke()

        # depends on border direction
        idx = 0 if direction=='row' else 1

        # cell range
        x0, x1 = self.merged_bbox[idx], self.merged_bbox[idx+2]

        # check all candidate strokes
        L = 0.0
        border_strokes = []
        for stroke in strokes:
            bbox = (stroke.x0, stroke.y0, stroke.x1, stroke.y1)
            t0, t1 = bbox[idx], bbox[idx+2]
            if t1 <= x0: continue
            if t0 >= x1: break
            # intersection length
            dl = min(x1, t1) - max(x0, t0)
            # NOTE to ignore small intersection on end point
            if dl < constants.MAJOR_DIST: continue 
            L += dl
            border_strokes.append(stroke)
        
        # use an empty stroke if nothing found, especially for merged cells.
        # no worry since the border style will be set correctly by adjacent separate cells.
        if L/(x1-x0) < constants.FACTOR_MAJOR: return Stroke()

        # the entire border of merged cell must have same property, i.e. width and color
        # otherwise, set empty stroke here and let adjacent cells set the correct style separately
        if len(border_strokes)==1: return border_strokes[0]

        properties = set([stroke.color for stroke in border_strokes])
        return border_strokes[0] if len(properties)==1 else Stroke()


class TableStructure:
    '''Parsing table structure based on strokes/fills.
    
    Steps to parse table structure::

            x0        x1       x2        x3
        y0  +----h1---+---h2---+----h3---+
            |         |        |         |
            v1        v2       v3        v4
            |         |        |         |
        y1  +----h4------------+----h5---+
            |                  |         |
            v5                 v6        v7
            |                  |         |
        y2  +--------h6--------+----h7---+
        

    1. Group horizontal and vertical strokes::
        
        self.h_strokes = {
            y0 : [h1, h2, h3],
            y1 : [h4, h5],
            y2 : [h6, h7]
        }
    
    These ``[x0, x1, x2, x3] x [y0, y1, y2]`` forms table lattices, i.e. 2 rows x 3 cols.

    2. Check merged cells in row/column direction.

    Let horizontal line ``y=(y0+y1)/2`` cross through table, it gets intersection with 
    ``v1``, ``v2`` and ``v3``, indicating no merging exists for cells in the first row.

    When ``y=(y1+y2)/2``, it has no intersection with vertical strokes at ``x=x1``, i.e. 
    merging status is ``[1, 0, 1]``, indicating ``Cell(2,2)`` is merged into ``Cell(2,1)``.

    So, the final merging status in this case::

        [
            [(1,1), (1,1), (1,1)],
            [(1,2), (0,0), (1,1)]
        ]
    '''

    def __init__(self, strokes:Shapes, **settings):
        '''Parse table structure from strokes and fills shapes.
        
        Args:
            strokes (Shapes): Stroke shapes representing table border. 
                For lattice table, they're retrieved from PDF raw contents; 
                for stream table, they're determined from layout of text blocks.

        .. note::
            Strokes must be sorted in reading order in advance, required by checking merged cells.        
        '''
        # cells
        self.cells = [] # type: list[list[CellStructure]]

        # group horizontal/vertical strokes -> table structure dict
        self.h_strokes, self.v_strokes = TableStructure._group_h_v_strokes(strokes, 
                        settings['min_border_clearance'], 
                        settings['max_border_width'])
        if not self.h_strokes or not self.v_strokes: return

        # initialize cells
        self.cells = self._init_cells()
        
    
    @property
    def bbox(self):
        '''Table boundary bbox.

        Returns:
            fitz.Rect: bbox of table.
        '''
        if not self.cells: return fitz.Rect()
        x0, y0 = self.cells[0][0].bbox.tl
        x1, y1 = self.cells[-1][-1].bbox.br
        return fitz.Rect(x0,y0,x1,y1)

    @property
    def num_rows(self): return len(self.cells)

    @property
    def num_cols(self): return len(self.cells[0]) if self.cells else 0

    @property
    def y_rows(self):
        """Top y-coordinate ``y0`` of each row.

        Returns:
            list: y-coordinates of each row.
        """        
        if not self.cells: return []
        Y = [row[0].bbox.y0 for row in self.cells]
        Y.append(self.cells[-1][0].bbox.y1)
        return Y

    @property
    def x_cols(self):
        """Left x-coordinate ``x0`` of each column.

        Returns:
            list: x-coordinates of each column.
        """  
        if not self.cells: return []
        X = [cell.bbox.x0 for cell in self.cells[0]]
        X.append(self.cells[0][-1].bbox.x1)
        return X


    def parse(self, fills:Shapes):
        '''Parse table structure.
        
        Args:
            fills (Shapes): Fill shapes representing cell shading.
        '''
        if not self.cells: return self

        # check merged cells
        self._check_merging_status()

        # check cell borders/shadings
        for row in self.cells:
            for cell in row:
                if cell.is_merged: continue
                cell.parse_borders(self.h_strokes, self.v_strokes)
                cell.parse_shading(fills)
        
        return self

    
    def to_table_block(self):
        '''Convert parsed table structure to ``TableBlock`` instance.

        Returns:
            TableBlock: Parsed table block instance.
        '''
        table = TableBlock()
        for row_structures in self.cells:
            # row object
            row = Row()
            row.height = row_structures[0].bbox.y1-row_structures[0].bbox.y0
            for cell_structure in row_structures:
                # if current cell is merged horizontally or vertically, set None.
                # actually, it will be counted in the top-left cell of the merged range.
                if cell_structure.is_merged:
                    row.append(Cell())
                    continue

                # cell borders properties
                top, bottom, left, right = cell_structure.borders
                w_top = top.width
                w_right = right.width
                w_bottom = bottom.width
                w_left = left.width

                # cell bg-color
                bg_color = cell_structure.shading.color if cell_structure.shading else None

                # Cell object
                # Note that cell bbox is calculated under real page CS, so needn't to consider rotation.
                cell = Cell({
                    'bg_color':  bg_color,
                    'border_color': (top.color, right.color, bottom.color, left.color),
                    'border_width': (w_top, w_right, w_bottom, w_left),
                    'merged_cells': cell_structure.merged_cells,
                }).update_bbox(cell_structure.merged_bbox)

                # add cell to row
                row.append(cell)                    
            
            # add row to table
            table.append(row)

        # finalize table structure
        if table: self._finalize_strokes_fills()

        return table


    def _finalize_strokes_fills(self):
        '''Finalize table structure, so set strokes and fills type as BORDER and SHADING accordingly.'''
        # strokes -> borders
        for k, strokes in self.h_strokes.items():
            for stroke in strokes: stroke.type = RectType.BORDER
        
        for k, strokes in self.v_strokes.items():
            for stroke in strokes: stroke.type = RectType.BORDER
        
        # fills -> shadings
        for row in self.cells:
            for cell in row:
                if cell.shading: cell.shading.type = RectType.SHADING


    @staticmethod
    def _group_h_v_strokes(strokes:Shapes, min_border_clearance:float, max_border_width:float):
        '''Split strokes in horizontal and vertical groups respectively.

        According to strokes below, the grouped h-strokes looks like::

            h_strokes = {
                y0 : [h1, h2, h3],
                y1 : [h4, h5],
                y2 : [h6, h7]
            }

               x0        x1        x2        x3
            y0  +----h1---+---h2---+----h3---+
                |         |        |         |
                v1        v2       v3        v4
                |         |        |         |
            y1  +----h4------------+----h5---+
                |                  |         |
                v5                 v6        v7
                |                  |         |
            y2  +--------h6--------+----h7---+

        '''
        def group_strokes(stroke:Shape, strokes:dict):
            # y-coordinate of h-strokes or x-coordinate of v-strokes
            t = round(stroke.y0, 1) if stroke.horizontal else round(stroke.x0, 1)

            # ignore minor error resulting from different stroke width
            for t_ in strokes:
                if abs(t-t_)>min_border_clearance: continue
                t = (t_+t)/2.0 # average
                strokes[t] = strokes.pop(t_)
                strokes[t].append(stroke)
                break
            else:
                strokes[t] = Shapes([stroke])
        
        h_strokes = {} # type: dict [float, Shapes]
        v_strokes = {} # type: dict [float, Shapes]
        X0, Y0, X1, Y1 = float('inf'), float('inf'), -float('inf'), -float('inf')
        for stroke in strokes:
            # group horizontal/vertical strokes in each row/column
            group_strokes(stroke, h_strokes if stroke.horizontal else v_strokes)

            # update table region
            X0 = min(X0, stroke.x0)
            X1 = max(X1, stroke.x1)
            Y0 = min(Y0, stroke.y0)
            Y1 = max(Y1, stroke.y1)

        # at least 2 inner strokes exist
        if not h_strokes or not v_strokes: return None, None

        # Note: add dummy strokes if no outer strokes exist        
        table_bbox = Element().update_bbox((X0, Y0, X1, Y1)) # table bbox
        TableStructure._check_outer_strokes(table_bbox, h_strokes, 'top', max_border_width)
        TableStructure._check_outer_strokes(table_bbox, h_strokes, 'bottom', max_border_width)
        TableStructure._check_outer_strokes(table_bbox, v_strokes, 'left', max_border_width)
        TableStructure._check_outer_strokes(table_bbox, v_strokes, 'right', max_border_width)

        # ATTENTION: sort in advance to avoid mistake when checking cell merging status
        for _, borders in h_strokes.items(): borders.sort_in_line_order()        
        for _, borders in v_strokes.items(): borders.sort_in_reading_order()

        return h_strokes, v_strokes
    

    def _init_cells(self):
        '''Initialize table lattices.'''
        # sort keys of borders -> table rows/cols coordinates
        y_rows = sorted(self.h_strokes)
        x_cols = sorted(self.v_strokes)

        # each lattice is a cell
        cells = []
        for i in range(len(y_rows)-1):
            y0, y1 = y_rows[i], y_rows[i+1]
            cells.append([]) # append a row
            for j in range(len(x_cols)-1):
                x0, x1 = x_cols[j], x_cols[j+1]
                cell = CellStructure([x0, y0, x1, y1])
                cells[-1].append(cell)
        
        return cells


    def _check_merging_status(self):
        '''Check cell merging status.'''
        x_cols, y_rows = self.x_cols, self.y_rows
        # check cell merging status in each row
        merged_cells_rows = []  # type: list[list[int]]
        ordered_strokes = [self.v_strokes[k] for k in x_cols]
        for row in self.cells:
            ref_y = (row[0].bbox.y0+row[0].bbox.y1)/2.0
            row_structure = TableStructure._check_merged_cells(ref_y, ordered_strokes, 'row')
            merged_cells_rows.append(row_structure)

        # check cell merging status in each column
        merged_cells_cols = []  # type: list[list[int]]
        ordered_strokes = [self.h_strokes[k] for k in y_rows]
        for cell in self.cells[0]:
            ref_x = (cell.bbox.x0+cell.bbox.x1)/2.0
            col_structure = TableStructure._check_merged_cells(ref_x, ordered_strokes, 'column')
            merged_cells_cols.append(col_structure)

        # count merged cells in row and column directions
        for i in range(self.num_rows):
            for j in range(self.num_cols):
                cell = self.cells[i][j]
                n_col = TableStructure._count_merged_cells(merged_cells_rows[i][j:])
                n_row = TableStructure._count_merged_cells(merged_cells_cols[j][i:])                
                cell.merged_cells = (n_row, n_col)        

        # check whether merged region is valid
        for i in range(self.num_rows):
            for j in range(self.num_cols):
                # validate region
                self._validate_merging_region(i, j)

                # update merged bbox
                # A separate cell without merging can also be treated as a merged range 
                # with 1 row and 1 colum, i.e. itself.
                cell = self.cells[i][j]               
                n_row, n_col = cell.merged_cells
                bbox = (x_cols[j], y_rows[i], x_cols[j+n_col], y_rows[i+n_row])
                cell.merged_bbox = fitz.Rect(bbox)

    
    @staticmethod
    def _check_outer_strokes(table_bbox:Element, borders:dict, direction:str, max_border_width:float):
        '''Add missing outer borders based on table bbox and grouped horizontal/vertical borders.
        
        Args:
            * table_bbox (Element): Table region.
            * borders (dict): Grouped horizontal (or vertical) borders at y-coordinates.
            * direction (str): Either ``top`` or ``bottom`` or ``left`` or ``right``.
        '''
        # target / real borders
        bbox = list(table_bbox.bbox)
        if direction=='top':
            idx = 1
            current = min(borders)
            borders[current].sort_in_line_order()
        elif direction=='bottom':
            idx = 3
            current = max(borders)
            borders[current].sort_in_line_order()
        elif direction=='left':
            idx = 0
            current = min(borders)
            borders[current].sort_in_reading_order()
        elif direction=='right':
            idx = 2
            current = max(borders)
            borders[current].sort_in_reading_order()
        else:
            return
        target = bbox[idx]
        
        # add missing border rects
        sample_border = Stroke()        
        bbox[idx] = target
        bbox[(idx+2)%4] = target

        # add whole border if not exist
        if abs(target-current)> max_border_width:
            borders[target] = Shapes([sample_border.copy().update_bbox(bbox)])
        
        # otherwise, check border segments
        else:
            idx_start = (idx+1)%2 # 0, 1
            idx_end = idx_start+2

            occupied = [(border.bbox[idx_start], 
                        border.bbox[idx_end]) for border in borders[current]]
            occupied.append((bbox[idx_end], None)) # end point
            start = bbox[idx_start] # start point
            segments = []
            for (left, right) in occupied:
                end = left
                # not connected -> add missing border segment
                if abs(start-end)>constants.MINOR_DIST:
                    bbox[idx_start] = start
                    bbox[idx_start+2] = end
                    segments.append(sample_border.copy().update_bbox(bbox))
                
                # update ref position
                start = right
            
            borders[current].extend(segments)


    @staticmethod
    def _check_merged_cells(ref:float, borders:list, direction:str='row'):
        '''Check merged cells in a row/column. 
        
        Args:
            * ref (float): y (or x) coordinate of horizontal (or vertical) passing-through line.
            * borders (list[Shapes]): A list of vertical (or horizontal) rects list in a column (or row).
            * direction (str): ``row`` - check merged cells in row; ``column`` - check merged cells in a column.

        Taking cells in a row for example, give a horizontal line ``y=ref`` passing through this row, 
        check the intersection with vertical borders. The ``n-th`` cell is merged if no intersection 
        with the ``n-th`` border.
            
                +-----+-----+-----+
                |     |     |     |
                |     |     |     |
                +-----+-----------+
                |           |     |
            ----1-----0-----1----------> [1,0,1]
                |           |     |
                |           |     |
                +-----------+-----+
        '''
        res = [] # type: list[int]
        for shapes in borders[0:-1]:
            # NOTE: shapes MUST be sorted in reading order!!
            # multi-lines exist in a row/column
            for border in shapes:
                # reference coordinates depending on checking direction
                if direction=='row':
                    ref0, ref1 = border.y0, border.y1
                else:
                    ref0, ref1 = border.x0, border.x1

                # 1) intersection found
                if ref0 < ref < ref1:
                    res.append(1)
                    break
                
                # 2) reference line locates below current border:
                # still have a chance to find intersection with next border, but,
                # no chance if this is the last border, see the else-clause
                elif ref > ref1:
                    continue

                # 3) current border locates below the reference line:
                # no intersection is possible any more
                elif ref < ref0:
                    res.append(0)
                    break
            
            # see notes 2), no change any more
            else:
                res.append(0)

        return res


    @staticmethod
    def _count_merged_cells(merging_status:list):
        '''Count merged cells, 
        e.g. ``[1,0,0,1]`` -> the second and third cells are merged into the first one.
        
        Args:
            merging_status (list): A list of 0-1 representing cell merging status.
        '''
        # it's merged by other cell
        if merging_status[0]==0: return 0
        
        # check a continuous sequence of 0 status
        num = 1
        for val in merging_status[1:]:
            if val==0:
                num += 1
            else: 
                break            
        return num
    

    def _validate_merging_region(self, i:int, j:int):
        '''Check whether the merging region of Cell (i,j) is valid. If not, unset merging status. 

        Args:
            i (int): Row index of the target cell.
            j (int): Column index of the target cell.
        '''
        cell = self.cells[i][j]
        if cell.is_merged: return

        # merged cells count
        n_row, n_col = cell.merged_cells 
        if n_row==1 and n_col==1: return

        # unset merging status if invalid merging region
        if not self._is_valid_region(i, i+n_row, j, j+n_col):
            for m in range(i, i+n_row):
                for n in range(j, j+n_col):
                    target = self.cells[m][n]
                    if target.is_merged: target.merged_cells = (1, 1)
            # reset current cell
            cell.merged_cells = (1, 1)        


    def _is_valid_region(self, row_start:int, row_end:int, col_start:int, col_end:int):
        '''Check whether all cells in given region are marked to merge.

        Args:
            row_start (int): Start row index (included) of the target region.
            row_end (int): End row index (excluded) of the target region.
            col_start (int): Start column index (included) of the target region.
            col_end (int): Start column index (excluded) of the target region.
        '''
        for i in range(row_start, row_end):
            for j in range(col_start, col_end):
                if i==row_start and j==col_start: continue # skip top-left cell
                if not self.cells[i][j].is_merged:
                    return False
        return True




# -*- coding: utf-8 -*-


# -------------------------------------
# basic unit
# -------------------------------------
PT = 1.0   # basic unit
ITP = 72.0 # inch to point

MAJOR_DIST = 5.0 * PT   # significant distance exists between two block lines
MINOR_DIST = 1.0 * PT   # small distance
TINY_DIST  = 0.5 * PT   # very small distance

FACTOR_SAME = 0.99
FACTOR_ALMOST = 0.95
FACTOR_MOST = 0.90
FACTOR_MAJOR = 0.75
FACTOR_A_HALF = 0.5
FACTOR_A_FEW = 0.1
FACTOR_FEW = 0.01


# -------------------------------------
# docx
# -------------------------------------
HIDDEN_W_BORDER = 0.0   # do not show border
MIN_LINE_SPACING = 0.7  # minimum line spacing available in MS word
DEFAULT_LINE_SPACING = 1.02

# punctuation implying end of a sentense
SENTENCE_END_PUNC = '.．。?？!！'

# control characters not supported by lxml
# https://github.com/dothinking/pdf2docx/issues/126#issuecomment-1040034077
INVALID_CHARS = '\x00\x01\x02\x03\x04\x05\x06\x07\x08\x0b\x0c\x0e\x0f\x10\x11\x12\x13\x14\x15\x16\x17\x18\x19\x1a\x1b\x1c\x1d\x1e\x1f'


# -------------------------------------
# font parameters
# -------------------------------------

# CJK font definition parameters
CJK_CODEPAGE_BITS = {
    "JIS/Japan": 17,
    "Chinese: Simplified chars—PRC and Singapore": 18,
    "Korean Wansung": 19,
    "Chinese: Traditional chars—Taiwan and Hong Kong": 20,
    "Korean Johab": 21
}

CJK_UNICODE_RANGE_BITS = {
    'Hangul Jamo': 28,
    'Hiragana': 49,
    'Katakana': 50,
    'Bopomofo': 51,
    'Hangul Compatibility Jamo': 52,
    'Enclosed CJK Letters And Months': 54,
    'CJK Compatibility': 55,
    'Hangul Syllables': 56,
    'CJK Unified Ideographs': 59,
    'CJK Strokes': 61,
    'Yi Syllables': 83
}

CJK_UNICODE_RANGES = [
    [0x1100, 0x11FF],  # Hangul Jamo
    [0x3040, 0x309F],  # Hiragana
    [0x30A0, 0x30FF],  # Katakana
    [0x31F0, 0x31FF],  # Katakana Phonetic Extensions
    [0x3100, 0x312F],  # Bopomofo
    [0x31A0, 0x31BF],  # Bopomofo Extended (Bopomofo)
    [0x3130, 0x318F],  # Hangul Compatibility Jamo
    [0x3200, 0x32FF],  # Enclosed CJK Letters and Months
    [0x3300, 0x33FF],  # CJK Compatibility
    [0xAC00, 0xD7AF],  # Hangul Syllables
    [0x4E00, 0x9FFF],  # CJK Unified Ideographs
    [0x2E80, 0x2EFF],  # CJK Radicals Supplement (CJK Unified Ideographs)
    [0x2F00, 0x2FDF],  # Kangxi Radicals (CJK Unified Ideographs)
    [0x2FF0, 0x2FFF],  # Ideographic Description Characters (CJK Unified Ideographs)
    [0x3400, 0x4DBF],  # CJK Unified Ideographs Extension A (CJK Unified Ideographs)
    [0x20000, 0x2A6DF],  # CJK Unified Ideographs Extension B (CJK Unified Ideographs)
    [0x3190, 0x319F],  # Kanbun (CJK Unified Ideographs)
    [0x31C0, 0x31EF],  # CJK Strokes
    [0xF900, 0xFAFF],  # CJK Compatibility Ideographs (CJK Strokes)
    [0x2F800, 0x2FA1F],  # CJK Compatibility Ideographs Supplement (CJK Strokes)
    [0xA000, 0xA48F],  # Yi Syllables
    [0xA490, 0xA4CF],  # Yi Radicals
]

DEFAULT_FONT_NAME = 'Arial'


# -*- coding: utf-8 -*-

'''Section of Page.

In most cases, one section per page. But in case multi-columns page, sections are used 
to distinguish these different layouts.

.. note::
    Currently, support at most two columns.

::

    {
        'bbox': (x0,y0,x1,y1)
        'num_cols': 1,
        'space': 0,
        'columns': [{
            ... # column properties
        }, ...]
    }
'''

from docx.enum.section import WD_SECTION
from ..common.docx import set_columns
from ..common.Collection import BaseCollection
from .Column import Column


class Section(BaseCollection):
    
    def __init__(self, space:int=0, columns:list=None, parent=None):
        """Initialize Section instance.

        Args:
            space (int, optional): Space between adjacent columns. Defaults to 0.
            columns (list, optional): A list of Column instances. Defaults to None.
            parent (Sections, optional): Parent element. Defaults to None.
        """
        self.space = space
        self.before_space = 0.0
        super().__init__(columns, parent)
    

    @property
    def num_cols(self): return len(self)


    def store(self):
        '''Store parsed section layout in dict format.'''
        return {
            'bbox'   : tuple([x for x in self.bbox]),
            'num_cols'   : self.num_cols,
            'space'  : self.space,
            'before_space'  : self.before_space,
            'columns': super().store()
        }
    

    def restore(self, raw:dict):
        '''Restore section from source dict.'''
        # bbox is maintained automatically based on columns
        self.space = raw.get('space', 0)   # space between adjacent columns
        self.before_space = raw.get('before_space', 0)   # space between adjacent columns

        # get each column
        for raw_col in raw.get('columns', []):
            column = Column().restore(raw_col)
            self.append(column)

        return self


    def parse(self, **settings):
        '''Parse section layout.'''
        for column in self: column.parse(**settings)        
        return self
    

    def make_docx(self, doc):
        '''Create section in docx. 

        Args:
            doc (Document): ``python-docx`` document object
        '''
        # set section column
        section = doc.sections[-1]
        width_list = [c.bbox[2]-c.bbox[0] for c in self]
        set_columns(section, width_list, self.space)

        # add create each column
        for column in self:
            # column break to start new column
            if column != self[0]: 
                doc.add_section(WD_SECTION.NEW_COLUMN)

            # make doc
            column.make_docx(doc)


# -*- coding: utf-8 -*-

'''A wrapper of pdf page engine (e.g. PyMuPDF, pdfminer) to do the following work:

* extract source contents
* clean up blocks/shapes, e.g. elements out of page
* calculate page margin
* parse page structure roughly, i.e. section and column
'''

from .BasePage import BasePage
from ..layout.Layout import Layout
from ..layout.Section import Section
from ..layout.Column import Column
from ..shape.Shape import Hyperlink
from ..shape.Shapes import Shapes
from ..font.Fonts import Fonts
from ..text.TextSpan import TextSpan
from ..common.share import debug_plot
from ..common import constants
from ..common.Collection import Collection


class RawPage(BasePage, Layout):
    '''A wrapper of page engine.'''

    def __init__(self, page_engine=None):
        ''' Initialize page layout.
        
        Args:
            page_engine (Object): Source pdf page.
        '''
        BasePage.__init__(self)
        Layout.__init__(self)
        self.page_engine = page_engine
    

    def extract_raw_dict(self, **settings):
        '''Extract source data with page engine. Return a dict with the following structure:
        ```
            {
                "width" : w,
                "height": h,    
                "blocks": [{...}, {...}, ...],
                "shapes" : [{...}, {...}, ...]
            }
        ```
        '''
        raise NotImplementedError
    
    @property
    def text(self):
        '''All extracted text in this page, with images considered as ``<image>``. 
        Should be run after ``restore()`` data.'''
        return '\n'.join([block.text for block in self.blocks])

    @property
    def raw_text(self):
        '''Extracted raw text in current page. Should be run after ``restore()`` data.'''
        return '\n'.join([block.raw_text for block in self.blocks])


    @debug_plot('Source Text Blocks')
    def restore(self, **settings):
        '''Initialize layout extracted with ``PyMuPDF``.'''
        raw_dict = self.extract_raw_dict(**settings)
        super().restore(raw_dict)
        return self.blocks

    
    @debug_plot('Cleaned Shapes')
    def clean_up(self, **settings):
        '''Clean up raw blocks and shapes, e.g. 
        
        * remove negative or duplicated instances,
        * detect semantic type of shapes
        '''
        # clean up blocks first
        self.blocks.clean_up(
            settings['float_image_ignorable_gap'],
            settings['line_overlap_threshold'])

        # clean up shapes        
        self.shapes.clean_up(
            settings['max_border_width'],
            settings['shape_min_dimension'])
        
        return self.shapes


    def process_font(self, fonts:Fonts):      
        '''Update font properties, e.g. font name, font line height ratio, of ``TextSpan``.
        
        Args:
            fonts (Fonts): Fonts parsed by ``fonttools``.
        '''
        # get all text span
        spans = []
        for line in self.blocks:
            spans.extend([span for span in line.spans if isinstance(span, TextSpan)])

        # check and update font name, line height
        for span in spans:
            font = fonts.get(span.font)
            if not font: continue

            # update font properties with font parsed by fonttools
            span.font = font.name
            if font.line_height:
                span.line_height = font.line_height * span.size


    def calculate_margin(self, **settings):
        """Calculate page margin.

        .. note::
            Ensure this method is run right after cleaning up the layout, so the page margin is 
            calculated based on valid layout, and stay constant.
        """
        # Exclude hyperlink from shapes because hyperlink might exist out of page unreasonably, 
        # while it should always within page since attached to text.
        shapes = Shapes([shape for shape in self.shapes if not isinstance(shape, Hyperlink)])

        # return default margin if no blocks exist
        if not self.blocks and not shapes: return (constants.ITP, ) * 4

        x0, y0, x1, y1 = self.bbox
        u0, v0, u1, v1 = self.blocks.bbox | shapes.bbox

        # margin
        left = max(u0-x0, 0.0)
        right = max(x1-u1-constants.MINOR_DIST, 0.0)
        top = max(v0-y0, 0.0)
        bottom = max(y1-v1, 0.0)

        # reduce calculated top/bottom margin to leave some free space
        top *= settings['page_margin_factor_top']
        bottom *= settings['page_margin_factor_bottom']

        # use normal margin if calculated margin is large enough
        return (
            min(constants.ITP, round(left, 1)), 
            min(constants.ITP, round(right, 1)), 
            min(constants.ITP, round(top, 1)), 
            min(constants.ITP, round(bottom, 1)))


    def parse_section(self, **settings):
        '''Detect and create page sections.

        .. note::
            - Only two-columns Sections are considered for now.
            - Page margin must be parsed before this step.
        '''
        # bbox
        X0, Y0, X1, _ = self.working_bbox
    
        # collect all blocks (line level) and shapes
        elements = Collection()
        elements.extend(self.blocks)
        elements.extend(self.shapes)
        if not elements: return

        # to create section with collected lines        
        lines = Collection()        
        sections = []
        def close_section(num_col, elements, y_ref):
            # append to last section if both single column
            if sections and sections[-1].num_cols==num_col==1:
                column = sections[-1][0] # type: Column
                column.union_bbox(elements)
                column.add_elements(elements)
            # otherwise, create new section
            else:
                section = self._create_section(num_col, elements, (X0, X1), y_ref)
                if section: 
                    sections.append(section)


        # check section row by row
        pre_num_col = 1
        y_ref = Y0 # to calculate v-distance between sections
        for row in elements.group_by_rows():
            # check column col by col
            cols = row.group_by_columns()
            current_num_col = len(cols)

            # column check:
            # consider 2-cols only
            if current_num_col>2:
                current_num_col = 1
            
            # the width of two columns shouldn't have significant difference
            elif current_num_col==2:
                u0, v0, u1, v1 = cols[0].bbox
                m0, n0, m1, n1 = cols[1].bbox
                x0 = (u1+m0)/2.0
                c1, c2 = x0-X0, X1-x0 # column width
                w1, w2 = u1-u0, m1-m0 # line width
                f = 2.0
                if not 1/f<=c1/c2<=f or w1/c1<0.33 or w2/c2<0.33: 
                    current_num_col = 1

            # process exceptions
            if pre_num_col==2 and current_num_col==1:
                # though current row has one single column, it might have another virtual 
                # and empty column. If so, it should be counted as 2-cols
                cols = lines.group_by_columns()
                pos = cols[0].bbox[2]
                if row.bbox[2]<=pos or row.bbox[0]>pos:
                    current_num_col = 2
                
                # pre_num_col!=current_num_col => to close section with collected lines,
                # before that, further check the height of collected lines
                else:
                    x0, y0, x1, y1 = lines.bbox
                    if y1-y0<settings['min_section_height']:
                        pre_num_col = 1
                

            elif pre_num_col==2 and current_num_col==2:
                # though both 2-cols, they don't align with each other
                combine = Collection(lines)
                combine.extend(row)
                if len(combine.group_by_columns(sorted=False))==1: current_num_col = 1


            # finalize pre-section if different from the column count of previous section
            if current_num_col!=pre_num_col:
                # process pre-section
                close_section(pre_num_col, lines, y_ref)
                if sections: 
                    y_ref = sections[-1][-1].bbox[3]

                # start potential new section                
                lines = Collection(row)
                pre_num_col = current_num_col

            # otherwise, collect current lines for further processing
            else:
                lines.extend(row)

        # don't forget the final section
        close_section(current_num_col, lines, y_ref)

        return sections
    

    @staticmethod
    def _create_section(num_col:int, elements:Collection, h_range:tuple, y_ref:float):
        '''Create section based on column count, candidate elements and horizontal boundary.'''
        if not elements: return
        X0, X1 = h_range

        if num_col==1:
            x0, y0, x1, y1 = elements.bbox
            column = Column().update_bbox((X0, y0, X1, y1))
            column.add_elements(elements)
            section = Section(space=0, columns=[column])
            before_space = y0 - y_ref
        else:
            cols = elements.group_by_columns()
            u0, v0, u1, v1 = cols[0].bbox
            m0, n0, m1, n1 = cols[1].bbox
            u = (u1+m0)/2.0

            column_1 = Column().update_bbox((X0, v0, u, v1))
            column_1.add_elements(elements)

            column_2 = Column().update_bbox((u, n0, X1, n1))
            column_2.add_elements(elements)

            section = Section(space=0, columns=[column_1, column_2])
            before_space = v0 - y_ref

        section.before_space = round(before_space, 1)
        return section


# -*- coding: utf-8 -*-

'''Collection of :py:class:`~pdf2docx.page.Page` instances.'''

import logging

from .RawPageFactory import RawPageFactory
from ..common.Collection import BaseCollection
from ..font.Fonts import Fonts


class Pages(BaseCollection):
    '''A collection of ``Page``.'''

    def parse(self, fitz_doc, **settings):
        '''Analyze document structure, e.g. page section, header, footer.

        Args:
            fitz_doc (fitz.Document): ``PyMuPDF`` Document instance.
            settings (dict): Parsing parameters.
        '''
        # ---------------------------------------------
        # 0. extract fonts properties, especially line height ratio
        # ---------------------------------------------
        fonts = Fonts.extract(fitz_doc)

        # ---------------------------------------------
        # 1. extract and then clean up raw page
        # ---------------------------------------------
        pages, raw_pages = [], []
        words_found = False
        for page in self:
            if page.skip_parsing: continue

            # init and extract data from PDF
            raw_page = RawPageFactory.create(page_engine=fitz_doc[page.id], backend='PyMuPDF')
            raw_page.restore(**settings)

            # check if any words are extracted since scanned pdf may be directed
            if not words_found and raw_page.raw_text.strip():
                words_found = True

            # process blocks and shapes based on bbox
            raw_page.clean_up(**settings)

            # process font properties
            raw_page.process_font(fonts)            

            # after this step, we can get some basic properties
            # NOTE: floating images are detected when cleaning up blocks, so collect them here
            page.width = raw_page.width
            page.height = raw_page.height
            page.float_images.reset().extend(raw_page.blocks.floating_image_blocks)

            raw_pages.append(raw_page)
            pages.append(page)

        # show message if no words found
        if not words_found:
            logging.warning('Words count: 0. It might be a scanned pdf, which is not supported yet.')

        
        # ---------------------------------------------
        # 2. parse structure in document/pages level
        # ---------------------------------------------
        # NOTE: blocks structure might be changed in this step, e.g. promote page header/footer,
        # so blocks structure based process, e.g. calculating margin, parse section should be 
        # run after this step.
        header, footer = Pages._parse_document(raw_pages)


        # ---------------------------------------------
        # 3. parse structure in page level, e.g. page margin, section
        # ---------------------------------------------
        # parse sections
        for page, raw_page in zip(pages, raw_pages):
            # page margin
            margin = raw_page.calculate_margin(**settings)
            raw_page.margin = page.margin = margin

            # page section
            sections = raw_page.parse_section(**settings)
            page.sections.extend(sections)
    

    @staticmethod
    def _parse_document(raw_pages:list):
        '''Parse structure in document/pages level, e.g. header, footer'''
        # TODO
        return '', ''


# -*- coding: utf-8 -*-
from enum import Enum
import random
from collections.abc import Iterable
from fitz.utils import getColorList, getColorInfoList


class BlockType(Enum):
    '''Block types.'''
    UNDEFINED = -1
    TEXT = 0
    IMAGE = 1
    LATTICE_TABLE = 2
    STREAM_TABLE = 3
    FLOAT_IMAGE = 4


class RectType(Enum):
    '''Shape type in context.'''
    HIGHLIGHT = 1
    UNDERLINE = 1<<1
    STRIKE = 1<<2
    HYPERLINK = 1<<3
    BORDER = 1<<4
    SHADING = 1<<5


class TextDirection(Enum):
    '''Text direction.    
    * LEFT_RIGHT: from left to right within a line, and lines go from top to bottom
    * BOTTOM_TOP: from bottom to top within a line, and lines go from left to right
    * MIX       : a mixture if LEFT_RIGHT and BOTTOM_TOP
    * IGNORE    : neither LEFT_RIGHT nor BOTTOM_TOP
    '''
    IGNORE     = -1
    LEFT_RIGHT = 0
    BOTTOM_TOP = 1
    MIX        = 2


class TextAlignment(Enum):
    '''Text alignment.

    .. note::
        The difference between ``NONE`` and ``UNKNOWN``: 

        * NONE: none of left/right/center align -> need TAB stop
        * UNKNOWN: can't decide, e.g. single line only
    '''
    NONE    = -1
    UNKNOWN = 0 
    LEFT    = 1
    CENTER  = 2
    RIGHT   = 3
    JUSTIFY = 4


class IText:
    '''Text related interface considering text direction.'''
    @property
    def text_direction(self):
        '''Text direction is from left to right by default.'''
        return TextDirection.LEFT_RIGHT

    @property
    def is_horizontal_text(self):
        '''Check whether text direction is from left to right.'''
        return self.text_direction == TextDirection.LEFT_RIGHT or \
                self.text_direction == TextDirection.MIX

    @property
    def is_vertical_text(self):
        '''Check whether text direction is from bottom to top.'''
        return self.text_direction == TextDirection.BOTTOM_TOP

    @property
    def is_mix_text(self):
        '''Check whether text direction is either from left to 
        right or from bottom to top.'''
        return self.text_direction == TextDirection.MIX


class lazyproperty:
    '''Calculate only once and cache property value.'''
    def __init__(self, func):
        self.func = func

    def __get__(self, instance, cls):
        if instance is None:
            return self
        else:
            value = self.func(instance)
            setattr(instance, self.func.__name__, value)
            return value


# -------------------------
# methods
# -------------------------
def is_number(str_number):
    '''Whether can be converted to a float.'''
    try:
        float(str_number)
    except ValueError:
        return False
    else:
        return True


def flatten(items, klass):
    '''Yield items from any nested iterable.'''
    for item in items:
        if isinstance(item, Iterable) and not isinstance(item, klass):
            yield from flatten(item, klass)
        else:
            yield item

def lower_round(number:float, ndigits:int=0):
    '''Round number to lower bound with specified digits, e.g. lower_round(1.26, 1)=1.2'''
    n = 10.0**ndigits
    return int(n*number) / n


# -------------------------
# color methods
# -------------------------
def rgb_component_from_name(name:str=''):
    '''Get a named RGB color (or random color) from fitz predefined colors, e.g. 'red' -> (1.0,0.0,0.0).'''
    # get color index
    if name and name.upper() in getColorList():
        pos = getColorList().index(name.upper())
    else:
        pos = random.randint(0, len(getColorList())-1)
        
    c = getColorInfoList()[pos]
    return (c[1] / 255.0, c[2] / 255.0, c[3] / 255.0)


def rgb_component(srgb:int):
    '''srgb value to R,G,B components, e.g. 16711680 -> (255, 0, 0).
    
    Equal to PyMuPDF built-in method::

        [int(255*x) for x in fitz.sRGB_to_pdf(x)]
    '''
    # decimal to hex: 0x...
    s = hex(srgb)[2:].zfill(6)
    return [int(s[i:i+2], 16) for i in [0, 2, 4]]


def rgb_to_value(rgb:list):
    '''RGB components to decimal value, e.g. (1,0,0) -> 16711680.'''
    res = 0
    for (i,x) in enumerate(rgb):
        res += int(x*(16**2-1)) * 16**(4-2*i)
    return int(res)


def cmyk_to_rgb(c:float, m:float, y:float, k:float, cmyk_scale:float=100):
    '''CMYK components to GRB value.'''
    r = (1.0 - c / float(cmyk_scale)) * (1.0 - k / float(cmyk_scale))
    g = (1.0 - m / float(cmyk_scale)) * (1.0 - k / float(cmyk_scale))
    b = (1.0 - y / float(cmyk_scale)) * (1.0 - k / float(cmyk_scale))
    res = rgb_to_value([r, g, b]) # type: int
    return res


def rgb_value(components:list):
    '''Gray/RGB/CMYK mode components to color value.'''
    num = len(components)
    # CMYK mode
    if num==4:
        c, m, y, k = map(float, components)
        color = cmyk_to_rgb(c, m, y, k, cmyk_scale=1.0)
    # RGB mode
    elif num==3:
        r, g, b = map(float, components)
        color = rgb_to_value([r, g, b])
    # gray mode
    elif num==1:
        g = float(components[0])
        color = rgb_to_value([g,g,g])    
    else:
        color = 0

    return color


# -------------------------
# pdf plot
# -------------------------
def new_page(doc, width:float, height:float, title:str):
    '''Insert a new page with given title.
    
    Args:
        doc (fitz.Document): pdf document object.
        width (float): Page width.
        height (float): Page height.
        title (str): Page title shown in page.
    '''
    # insert a new page
    page = doc.new_page(width=width, height=height)    

    # plot title at the top-left corner
    gray = rgb_component_from_name('gray')
    page.insert_text((5, 16), title, color=gray, fontsize=15)
    
    return page


def debug_plot(title:str, show=True):
    '''Plot the returned objects of inner function.
    
    Args:
        title (str): Page title.
        show (bool, optional): Don't plot if show==False. Default to True.
    
    .. note::
        Prerequisite of the inner function: 
            - the first argument is a :py:class:`~pdf2docx.page.BasePage` instance.
            - the last argument is configuration parameters in ``dict`` type.
    '''
    def wrapper(func):
        def inner(*args, **kwargs):
            # execute function
            objects = func(*args, **kwargs)

            # check if plot page
            page = args[0] # BasePage object
            debug = kwargs.get('debug', False)
            doc = kwargs.get('debug_doc', None)
            filename = kwargs.get('debug_filename', None)

            if show and objects and debug and doc is not None:                
                # create a new page
                debug_page = new_page(doc, page.width, page.height, title)
                # plot objects, e.g. text blocks, shapes, tables...
                objects.plot(debug_page)
                doc.save(filename)

            return objects
        return inner
    return wrapper

